from typing import Tuple, Dict, Union
import numpy as np
from tqdm import tqdm
from scipy.stats import norm, poisson
import numba
import strax
import straxen
from straxen.plugins.defaults import DEFAULT_POSREC_ALGO, FAR_XYPOS_S2_TYPE, WIDE_XYPOS_S2_TYPE
from straxen.plugins.peaklets.peaklets import drop_data_field


export, __all__ = strax.exporter()


@export
class MergedS2s(strax.OverlapWindowPlugin):
    """Merge together peaklets if peak finding favours that they would form a single peak instead.
    Technically, the S2 merging algorithm merges S2 peaklets into S2 peaks. By introducing more
    information about the waveform and (x, y) distribution of potential groups of peaklets, the
    algorithm removes PI and DE population from S2 peaks.

    Note: Types FAR_XYPOS_S2_TYPE (20) and WIDE_XYPOS_S2_TYPE (22) are still S2s,
    but they do not participate in the event building.

    The algorithm merges S2 peaklets when they are close in (t, x, y). But if a group of peaklets
    is dense in time but sparse in (x, y), the following steps are conducted:

    1. Merge these peaklets that are dense in (x, y).
    2. Assign the peaklets that are dense in time but not dense in (x, y) by type 20,
        they are usually PI or DE.
    3. If the sum of nearby type 20 is large compared to the merged peak,
        assign the merged peak as type 22, because it is usually PI.

    Reference:
    xenon:xenonnt:analysis:s2_merging_time_position
    xenon:xenonnt:analysis:sr2_peak_types

    """

    __version__ = "2.0.1"

    depends_on: Tuple[str, ...] = (
        "peaklets",
        f"peaklet_positions_{DEFAULT_POSREC_ALGO}",
        "peaklet_classification",
        "lone_hits",
    )
    provides: Union[Tuple[str, ...], str] = ("merged_s2s", "enhanced_peaklet_classification")
    data_kind: Union[Dict[str, str], str] = dict(
        merged_s2s="merged_s2s", enhanced_peaklet_classification="peaklets"
    )

    n_tpc_pmts = straxen.URLConfig(type=int, help="Number of TPC PMTs")

    n_top_pmts = straxen.URLConfig(type=int, help="Number of top TPC array PMTs")

    default_reconstruction_algorithm = straxen.URLConfig(
        default=DEFAULT_POSREC_ALGO, help="default reconstruction algorithm that provides (x,y)"
    )

    s2_merge_max_duration = straxen.URLConfig(
        default=50_000,
        type=int,
        infer_type=False,
        help="Do not merge peaklets at all if the result would be a peak longer than this [ns]",
    )

    s2_merge_gap_thresholds = straxen.URLConfig(
        default=(
            (1.84, 2.84e04),
            (2.18, 2.40e04),
            (2.51, 1.96e04),
            (2.84, 1.80e04),
            (3.18, 1.68e04),
            (3.51, 1.86e04),
            (3.84, 1.98e04),
            (4.18, 1.66e04),
            (4.51, 1.21e04),
        ),
        infer_type=False,
        help=(
            "Points to define maximum separation between peaklets to allow "
            "merging [ns] depending on log10 area of the merged peak\n"
            "where the gap size of the first point is the maximum gap to allow merging."
            "The format is ((log10(area), max_gap), (..., ...), (..., ...))"
        ),
    )

    s2_merge_p_thresholds = straxen.URLConfig(
        default=(
            (4.18, 9.89e-04),
            (4.51, 7.15e-03),
            (4.84, 1.60e-01),
        ),
        infer_type=False,
        help=(
            "Points to define minimum p-value of merging proposal "
            "depending on log10 area of the merged peak\n"
            "The format is ((log10(area), p-value), (..., ...), (..., ...))"
        ),
    )

    s2_merge_dr_thresholds = straxen.URLConfig(
        default=(
            (1.51, 1.40e01),
            (1.84, 1.83e01),
            (2.18, 1.97e01),
            (2.51, 1.24e01),
            (2.84, 5.75e00),
        ),
        type=tuple,
        help=(
            "Points to define maximum weighted mean deviation of "
            "the peaklets from the main cluster [cm]\n"
            "The format is ((log10(area_top), dr), (..., ...), (..., ...))"
        ),
    )

    s2_merge_unmerged_thresholds = straxen.URLConfig(
        default=(1, 49.5, 0.01),
        type=tuple,
        help=(
            "Max (number, fraction of unmerged, total area of unmerged) "
            "of type 20 peaklets inside a peak. "
            "The number of type 20 peaklets should not be larger than the number threshold. "
            "The area of type 20 peaklets should not be larger than the both area thresholds. "
            "The fraction threshold is important when S2 is large, "
            "while the total area threshold is important when S2 is small."
        ),
    )

    merge_lone_hits = straxen.URLConfig(
        default=True,
        type=bool,
        help="Merge lone hits into merged S2s",
    )

    merge_s0 = straxen.URLConfig(
        default=True,
        type=bool,
        help="Merge S0s into merged S2s",
    )

    gain_model = straxen.URLConfig(
        infer_type=False,
        help="PMT gain model. Specify as (str(model_config), str(version), nT-->boolean",
    )

    rough_seg = straxen.URLConfig(
        default=30, type=(int, float), help="Rough single electron gain [PE/e]"
    )

    sigma_seg = straxen.URLConfig(
        default=6.5,
        type=(int, float),
        help="Standard deviation of the single electron gain [PE/e]",
    )

    rough_min_sigma = straxen.URLConfig(
        default=1e2, type=(int, float), help="Minimum sigma for the merged peaks"
    )

    rough_max_sigma = straxen.URLConfig(
        default=2e4, type=(int, float), help="Maximum sigma for the merged peaks"
    )

    rough_sigma_bins = straxen.URLConfig(
        default=10, type=int, help="Number of bins for sigma of merged peaks"
    )

    rough_mu_bins = straxen.URLConfig(
        default=10, type=int, help="Number of bins for mu of merged peaks"
    )

    poisson_max_mu = straxen.URLConfig(
        default=25, type=int, help="When to switch from Poisson to normal distribution"
    )

    poisson_survival_ratio = straxen.URLConfig(
        default=1e-4,
        type=float,
        help=(
            "Survival ratio for Poisson distribution. The PMF smaller than this will be ignored."
        ),
    )

    normal_max_sigma = straxen.URLConfig(
        default=7,
        type=(int, float),
        help="Maximum sigma for the normal distribution CDF panel",
    )

    normal_n_bins = straxen.URLConfig(
        default=501,
        type=int,
        help="Number of bins for the normal distribution CDF panel",
    )

    maxexp = straxen.URLConfig(
        default=10,
        type=(int, float),
        help="Maximum exponent for the posterior to keep numerical stability",
    )

    use_bayesian_merging = straxen.URLConfig(default=True, type=bool, help="Use Bayesian merging")

    rm_sparse_xy = straxen.URLConfig(
        default=True, type=bool, help="Remove peaklets that are too far away in (x, y)"
    )

    use_uncertainty_weights = straxen.URLConfig(
        default=True, type=bool, help="Use uncertainty from probabilistic posrec to derive weights"
    )

    p_value_prioritized = straxen.URLConfig(
        default=False,
        type=bool,
        help="Whether to prioritize p-value over area when testing the proposal",
    )

    merged_s2s_get_window_size_factor = straxen.URLConfig(
        default=5, type=int, track=False, help="Factor of the window size for the merged_s2s plugin"
    )

    disable_progress_bar = straxen.URLConfig(
        default=True, type=bool, track=False, help="Whether to disable the progress bar"
    )

    indicator_dtype = np.dtype(
        [(("Peaklet is merging input or peak is merged from peaklets", "merged"), bool)]
    )
    copied_dtype = np.dtype(
        [
            ("time", np.int64),
            ("endtime", np.int64),
            ("area", np.float32),
            ("median_time", np.float32),
            ("area_decile_from_midpoint", np.float32, (11,)),
        ]
    )

    def _have_data(self, field):
        return field in self.deps["peaklets"].dtype_for("peaklets").names

    def infer_dtype(self):
        peaklet_classification_dtype = self.deps["peaklet_classification"].dtype_for(
            "peaklet_classification"
        )
        peaklets_dtype = self.deps["peaklets"].dtype_for("peaklets")
        # The merged dtype is argument position dependent!
        # It must be first classification then peaklet
        # Otherwise strax will raise an error
        # when checking for the returned dtype!
        merged_s2s_dtype = strax.merged_dtype(
            (peaklet_classification_dtype, peaklets_dtype, self.indicator_dtype)
        )
        enhanced_peaklet_classification_dtype = strax.merged_dtype(
            (peaklet_classification_dtype, self.indicator_dtype)
        )
        return dict(
            merged_s2s=merged_s2s_dtype,
            enhanced_peaklet_classification=enhanced_peaklet_classification_dtype,
        )

    def setup(self):
        self.to_pe = self.gain_model
        self.gap_thresholds = np.array(self.s2_merge_gap_thresholds).T
        self.p_thresholds = np.array(self.s2_merge_p_thresholds).T
        self.dr_thresholds = np.array(self.s2_merge_dr_thresholds).T
        # Max gap and area should be set by the gap thresholds to avoid contradictions
        if np.argmax(self.gap_thresholds[1]) != 0:
            raise ValueError("The first point should be the maximum gap to allow merging")
        if self.p_thresholds[1].max() > 1 or self.p_thresholds[1].min() < 0:
            raise ValueError("P-value should be in the range of [0, 1]")
        self.max_gap = self.gap_thresholds[1, 0]
        self.max_duration = self.s2_merge_max_duration

        self.poisson_max_k = np.ceil(
            poisson.isf(q=self.poisson_survival_ratio, mu=self.poisson_max_mu)
        ).astype(int)
        self.factorial_panel = np.array(
            [np.prod(np.arange(1, k + 1, dtype=float)) for k in range(self.poisson_max_k + 1)]
        )
        if np.any(self.factorial_panel < 0):
            raise ValueError("Factorial panel has negative values, this might because of overflow")
        x = np.linspace(-self.normal_max_sigma, self.normal_max_sigma, self.normal_n_bins)
        self.normal_panel = np.vstack([x, norm.cdf(x)])

        self.sigma = np.linspace(self.rough_min_sigma, self.rough_max_sigma, self.rough_sigma_bins)
        self.sigma_panel = np.repeat(self.sigma[None, :], self.rough_mu_bins, axis=0).flatten()

        self.unmerged_thresholds = self.s2_merge_unmerged_thresholds

    def get_window_size(self):
        return self.merged_s2s_get_window_size_factor * (
            int(self.s2_merge_gap_thresholds[0][1]) + self.s2_merge_max_duration
        )

    def no_merging(self, peaklets):
        is_s2 = peaklets["type"] == 2
        return np.sum(is_s2) <= 1 or self.max_gap < 0

    def compute(self, peaklets, lone_hits, start, end):
        # "merged" will be set as False automatically
        _peaklets = strax.merge_arrs(
            [peaklets, np.zeros(len(peaklets), dtype=self.indicator_dtype)],
            dtype=strax.merged_dtype((peaklets.dtype, self.indicator_dtype)),
        )

        if self.use_uncertainty_weights:
            name = f"position_contour_{self.default_reconstruction_algorithm}"
            if name not in _peaklets.dtype.names:
                raise ValueError(f"{name} is not in the input peaklets dtype")

        # initialize enhanced_peaklet_classification
        # "merged" will be set as False automatically
        enhanced_peaklet_classification = np.zeros(
            len(_peaklets), dtype=self.dtype_for("enhanced_peaklet_classification")
        )
        # copy fields, especially type
        for d in enhanced_peaklet_classification.dtype.names:
            enhanced_peaklet_classification[d] = _peaklets[d]

        if self.no_merging(_peaklets):
            empty_result = self.empty_result()
            empty_result["enhanced_peaklet_classification"] = enhanced_peaklet_classification
            return empty_result

        # make sure the peaklets are not overwritten
        _peaklets.flags.writeable = False

        is_s2 = _peaklets["type"] == 2

        # peaklets might be overwritten in the merge method
        # so do not reuse the peaklets after this line
        merged_s2s, not_excluded, merged = self.merge(_peaklets, lone_hits, start, end)

        # mark the peaklets can be merged by time-density but not
        # by position-density as type FAR_XYPOS_S2_TYPE
        enhanced_peaklet_classification["type"][is_s2 & ~not_excluded] = FAR_XYPOS_S2_TYPE
        enhanced_peaklet_classification["merged"][merged] = True

        return dict(
            merged_s2s=merged_s2s, enhanced_peaklet_classification=enhanced_peaklet_classification
        )

    def merge(self, _peaklets, lone_hits, start, end):
        """Merge into S2s if the peaklets are close enough in time and position."""
        # this is an OverlapWindowPlugin, some peaklets will be reused in the next iteration
        # use _peaklets here to prevent peaklets from being overwritten
        if np.any(_peaklets["time"][1:] < strax.endtime(_peaklets)[:-1]):
            raise ValueError("Peaklets not disjoint, why?")

        # only keep S2 peaklets for merging
        is_s2 = _peaklets["type"] == 2

        if not (self._have_data("data_top") and self._have_data("data_start")):
            peaklets = np.zeros(
                is_s2.sum(),
                dtype=strax.merged_dtype(
                    (
                        strax.peak_dtype(
                            n_channels=self.n_tpc_pmts, store_data_top=True, store_data_start=True
                        ),
                        self.indicator_dtype,
                    )
                ),
            )
            strax.copy_to_buffer(_peaklets[is_s2], peaklets, "_add_data_top_or_start_field")
        else:
            peaklets = _peaklets[is_s2]

        # make sure the peaklets are not overwritten
        peaklets.flags.writeable = False

        max_buffer = int(self.max_duration // strax.gcd_of_array(peaklets["dt"]))

        # _not_excluded is the mask of the peaklets that are not excluded by (x, y) merging
        start_merge_at, end_merge_at, _not_excluded = self.get_merge_instructions(
            peaklets,
            start,
            end,
            self.max_gap,
            self.max_duration,
            self.sigma,
            self.rough_seg,
            self.sigma_seg,
            self.rough_mu_bins,
            self.poisson_max_mu,
            self.poisson_survival_ratio,
            self.normal_panel,
            self.factorial_panel,
            self.sigma_panel,
            self.maxexp,
            self.n_top_pmts,
            self.p_thresholds,
            self.dr_thresholds,
            self.default_reconstruction_algorithm,
            self.use_bayesian_merging,
            self.rm_sparse_xy,
            self.use_uncertainty_weights,
            self.p_value_prioritized,
            self.gap_thresholds,
            disable=self.disable_progress_bar,
        )

        if "data_top" not in peaklets.dtype.names or "data_start" not in peaklets.dtype.names:
            raise ValueError("data_top or data_start is not in the peaklets dtype")

        _merged_s2 = np.zeros(len(peaklets), dtype=bool)
        for i in range(len(start_merge_at)):
            sl = slice(start_merge_at[i], end_merge_at[i])
            _merged_s2[sl] = True
        # exclude the peaklets that are not merged by (x, y)
        _merged_s2[~_not_excluded] = False

        # have to redo the merging to prevent numerical instability
        is_s0 = _peaklets["type"] == 0
        if self.merge_s0 and len(start_merge_at) and is_s0.sum() > 0:
            # build the time interval of merged_s2s, even though they are not merged yet
            endtime = strax.endtime(peaklets)
            merged_s2s_window = np.zeros(len(start_merge_at), dtype=strax.time_fields)
            for i in range(len(start_merge_at)):
                sl = slice(start_merge_at[i], end_merge_at[i])
                merged_s2s_window["time"][i] = peaklets["time"][sl][_not_excluded[sl]][0]
                merged_s2s_window["endtime"][i] = endtime[sl][_not_excluded[sl]][-1]
            # the S0s that should be merged should fully be contained
            _merged_s0 = strax.fully_contained_in(_peaklets[is_s0], merged_s2s_window) != -1
            merged_s0s = strax.split_by_containment(_peaklets[is_s0], merged_s2s_window)
            # offsets of indices
            increments = np.array([len(m) for m in merged_s0s], dtype=int)
            offsets = np.hstack([0, np.cumsum(increments)])
            _start_merge_at = start_merge_at + offsets[:-1]
            _end_merge_at = end_merge_at + offsets[1:]
            if np.min(_end_merge_at - _start_merge_at) < 2:
                raise ValueError("You are merging nothing!")
            # prepare for peaklets including S0s
            __merged = np.hstack([_not_excluded, np.full(increments.sum(), True)])
            _peaklets = np.hstack([peaklets, np.hstack(merged_s0s)])
            argsort = strax.stable_argsort(_peaklets["time"])
            merged_s2s = self.merge_peaklets(
                _peaklets[argsort],
                _start_merge_at,
                _end_merge_at,
                __merged[argsort],
                max_buffer,
                max_unmerged=self.unmerged_thresholds,
            )
        else:
            _merged_s0 = np.zeros(is_s0.sum(), dtype=bool)
            merged_s2s = self.merge_peaklets(
                peaklets,
                start_merge_at,
                end_merge_at,
                _not_excluded,
                max_buffer,
                max_unmerged=self.unmerged_thresholds,
            )

        if (
            len(merged_s2s)
            and np.max((strax.endtime(merged_s2s) - merged_s2s["time"])) > self.max_duration
        ):
            raise ValueError("Merged S2 is too long")

        if self.merge_lone_hits:
            # Updated time and length of lone_hits and sort again:
            # this is an OverlapWindowPlugin, some lone_hits will be reused in the next iteration
            # so do not overwirte the lone_hits
            lh = np.copy(lone_hits)
            del lone_hits
            lh_time_shift = (lh["left"] - lh["left_integration"]) * lh["dt"]
            lh["time"] = lh["time"] - lh_time_shift
            lh["length"] = lh["right_integration"] - lh["left_integration"]
            lh = strax.sort_by_time(lh)

            _store_data_top = "data_top" in self.dtype_for("merged_s2s").names
            _store_data_start = "data_start" in self.dtype_for("merged_s2s").names
            strax.add_lone_hits(
                merged_s2s,
                lh,
                self.to_pe,
                n_top_channels=self.n_top_pmts,
                store_data_top=_store_data_top,
                store_data_start=_store_data_start,
            )

        strax.compute_properties(merged_s2s, n_top_channels=self.n_top_pmts)

        # remove position fields
        merged_s2s = drop_data_field(
            merged_s2s, self.dtype_for("merged_s2s"), "_drop_data_field_merged_s2s"
        )

        # make sure merged has same length as peaklets
        merged = np.zeros(len(is_s2), dtype=bool)
        merged[is_s2] = _merged_s2
        merged[is_s0] = _merged_s0
        not_excluded = np.ones(len(is_s2), dtype=bool)
        not_excluded[is_s2] = _not_excluded

        # of course all merged S2s are merged
        merged_s2s["merged"] = True

        return merged_s2s, not_excluded, merged

    @staticmethod
    def get_left_right(peaklet):
        """Get the left and right boundaries of the peaklet."""
        # The gap is defined as the 90% to 10% area decile distance of the adjacent peaks
        left = (peaklet["area_decile_from_midpoint"][1] + peaklet["median_time"]).astype(int)
        right = (peaklet["area_decile_from_midpoint"][9] + peaklet["median_time"]).astype(int)
        return left, right

    @staticmethod
    def get_gap(y, x):
        """Get the gap between two peaklets."""
        y_left, y_right = MergedS2s.get_left_right(y)
        x_left, x_right = MergedS2s.get_left_right(x)
        dt = x["time"] - y["time"]
        boundaries = np.sort([y_left, y_right, x_left + dt, x_right + dt])
        this_gap = boundaries[2] - boundaries[1]
        return this_gap

    @staticmethod
    @numba.njit(cache=True, nogil=True)
    def get_duration(y, x):
        """Get the duration of the merged peaklets."""
        return max(y["endtime"], x["endtime"]) - min(y["time"], x["time"])

    @staticmethod
    @numba.njit(cache=True, nogil=True)
    def decile_interp(
        x,
        y,
        panel=np.linspace(0, 1, 11),
    ):
        area = x["area"] + y["area"]
        xf = x["area"] / area
        yf = y["area"] / area
        xp = np.concatenate((panel * xf, panel * yf + xf))
        yp = np.concatenate(
            (
                (x["area_decile_from_midpoint"] + x["median_time"]),
                (y["time"] - x["time"]) + (y["area_decile_from_midpoint"] + y["median_time"]),
            )
        )
        decile = np.interp(panel, xp, yp)
        return decile[5], decile - decile[5]

    @staticmethod
    def merge_two_peaks(x, y):
        """Merge two peaklets in order."""
        z = np.array(0, dtype=x.dtype)
        z["time"] = x["time"]
        z["endtime"] = y["endtime"]
        z["area"] = x["area"] + y["area"]
        z["median_time"], z["area_decile_from_midpoint"] = MergedS2s.decile_interp(x, y)
        return z

    @staticmethod
    def get_merge_instructions(
        _peaks,
        start,
        end,
        max_gap,
        max_duration,
        sigma,
        rough_seg,
        sigma_seg,
        rough_mu_bins,
        poisson_max_mu,
        poisson_survival_ratio,
        normal_panel,
        factorial_panel,
        sigma_panel,
        maxexp,
        n_top_pmts,
        p_thresholds,
        dr_thresholds,
        posrec_algo,
        bayesian=True,
        sparse_xy=True,
        uncertainty_weights=True,
        p_value_prioritized=False,
        gap_thresholds=None,
        diagnosing=False,
        disable=True,
    ):
        """Find the group of peaklets to merge.

        There are two ways to merge peaklets:
        1. Bayesian merging: merge peaklets based on the p-value of the time-density merging
        2. Normal merging: merge peaklets based on the gap between the peaklets

        """

        # (x, y) positions of the peaklets
        positions = np.vstack([_peaks[f"x_{posrec_algo}"], _peaks[f"y_{posrec_algo}"]]).T
        if uncertainty_weights:
            contour_area = _peaks[f"position_contour_area_{posrec_algo}"]
        # weights of the peaklets when calculating the weighted mean deviation in (x, y)
        area = _peaks["area"]
        area_top = area * _peaks["area_fraction_top"]

        peaks = np.zeros(len(_peaks), dtype=MergedS2s.copied_dtype)
        for name in peaks.dtype.names:
            if name == "endtime":
                peaks[name] = strax.endtime(_peaks)
            else:
                peaks[name] = _peaks[name]
        peaks_copy = np.copy(peaks)
        peaks_copy.flags.writeable = False

        n_peaks = len(peaks)
        if n_peaks == 0:
            raise ValueError("No peaklets to merge")

        start_index = np.arange(n_peaks)
        # exclusive end index
        end_index = np.arange(n_peaks) + 1
        # keep a copy of the original indices
        start_index_copy = np.copy(start_index)
        start_index_copy.flags.writeable = False
        end_index_copy = np.copy(end_index)
        end_index_copy.flags.writeable = False

        # mask to help keep track of the peaklets that have been examined
        unexamined = np.full(n_peaks, True)
        # mask to help keep track of the peaklets that should not be merged because
        # of too high standard deviation from the main cluster in (x, y) of the peaklets
        merged = np.full(n_peaks, True)

        # approximation of the integration boundaries
        core_bounds = (peaks["time"][1:] + peaks["endtime"][:-1]) // 2
        # here the constraint on boundaries is also to make sure get_window_size covers the gaps
        left_bounds = np.maximum(np.hstack([start, core_bounds]), peaks["time"] - int(max_gap / 2))
        right_bounds = np.minimum(
            np.hstack([core_bounds, end]), peaks["endtime"] + int(max_gap / 2)
        )
        # keep a copy of the original boundaries
        left_bounds_copy = np.copy(left_bounds)
        left_bounds_copy.flags.writeable = False
        right_bounds_copy = np.copy(right_bounds)
        right_bounds_copy.flags.writeable = False

        if diagnosing:
            merged_area = []
            p_values = []
            dr_avgs = []

        def gap_small_enough(y, x):
            log_area = np.log10(y["area"] + x["area"])
            this_gap = MergedS2s.get_gap(y, x)
            gap_threshold = thresholds_interpolation(log_area, gap_thresholds)
            # gap of 90-10% area decile distance should not be larger than the threshold
            return this_gap < gap_threshold

        def can_merge_left(i, peaks):
            # whether peaklet i (from _peak) can be merged to its left unexamined peaklet
            flag = (
                i - 1 >= 0
                and unexamined[i - 1]
                and MergedS2s.get_gap(peaks[i], peaks[i - 1]) < max_gap
                and MergedS2s.get_duration(peaks[i], peaks[i - 1]) <= max_duration
            )
            if not flag:
                return False
            return gap_small_enough(peaks[i], peaks[i - 1])

        def can_merge_right(i, peaks):
            # whether peaklet i (from _peak) can be merged to its right unexamined peaklet
            flag = (
                i < n_peaks
                and unexamined[i]
                and MergedS2s.get_gap(peaks[i - 1], peaks[i]) < max_gap
                and MergedS2s.get_duration(peaks[i - 1], peaks[i]) <= max_duration
            )
            if not flag:
                return False
            return gap_small_enough(peaks[i - 1], peaks[i])

        argsort = strax.stable_argsort(peaks["area"])
        for i in tqdm(argsort[::-1], disable=disable):
            if not unexamined[i]:
                continue
            p = np.array([1.0, 1.0])
            p_threshold = np.array([0.0, 0.0])
            left = True
            dr_avg = 0.0
            # in the while loops, the peaklets will be merged until the p-value
            # is smaller than the threshold or the peaklets can not be merged anymore
            # i will NOT be updated in the while loop
            while np.any(p >= p_threshold) and unexamined[i]:
                indices = []
                # please mind here that the definition of gaps should
                # be consistent with in the merging algorithm
                # and the merged peak should not be longer than the max duration
                if can_merge_left(start_index[i], peaks):
                    indices.append(start_index[i] - 1)
                else:
                    indices.append(None)
                if can_merge_right(end_index[i], peaks):
                    indices.append(end_index[i])
                else:
                    indices.append(None)
                p = []
                p_threshold = []
                _area = []
                # get p-values
                for j in indices:
                    if j is None:
                        p.append(-1.0)
                        p_threshold.append(1.0)
                        _area.append(-1.0)
                        continue
                    # if area is non-positive, "merge" in time
                    # but still skip it later in (x, y) because its weight is nan
                    if not peaks["area"][j] > 0:
                        p.append(1.0)
                        p_threshold.append(-1.0)
                        _area.append(peaks["area"][j])
                        continue
                    log_area = np.log10(peaks["area"][i] + peaks["area"][j])
                    if bayesian:
                        p_ = get_p_value(
                            peaks[i],
                            peaks[j],
                            left_bounds[i],
                            right_bounds[i],
                            left_bounds[j],
                            right_bounds[j],
                            sigma,
                            rough_seg,
                            sigma_seg,
                            rough_mu_bins,
                            poisson_max_mu,
                            poisson_survival_ratio,
                            normal_panel,
                            factorial_panel,
                            sigma_panel,
                            maxexp,
                        )
                        p_threshold_ = thresholds_interpolation(
                            log_area,
                            p_thresholds,
                        )
                    else:
                        # this is kept for diagnosing and merged_s2s_he
                        p_ = 1.0
                        p_threshold_ = 0.0
                    p.append(p_)
                    p_threshold.append(p_threshold_)
                    _area.append(peaks["area"][j])
                p = np.array(p)
                p_threshold = np.array(p_threshold)
                _area = np.array(_area)
                examined = slice(start_index[i], end_index[i])

                if diagnosing:
                    merged_area.append(area[examined][merged[examined]].sum())

                if np.all(p < p_threshold):
                    # this will not allow merging of the already examined peaklets
                    unexamined[examined] = False
                else:
                    # whether test the proposal with larger area first
                    if p_value_prioritized:
                        left = p[0] > p[1]
                    else:
                        left = _area[0] > _area[1]
                    # if test left first
                    if left:
                        if p[0] >= p_threshold[0]:
                            left = True
                        else:
                            assert p[1] >= p_threshold[1]
                            left = False
                    else:
                        if p[1] >= p_threshold[1]:
                            left = False
                        else:
                            assert p[0] >= p_threshold[0]
                            left = True
                    # slice indicating the direction of merging
                    if left:
                        direction = [indices[0], indices[0] + 1]
                        index = indices[0]
                    else:
                        direction = [indices[1] - 1, indices[1]]
                        index = indices[1]
                    # slice indicating the peaklets to be merged
                    start_idx = start_index[direction[0]]
                    end_idx = end_index[direction[1]]
                    merging = slice(start_idx, end_idx)

                    if sparse_xy:
                        # calculate weighted averaged deviation of peaklets from the main cluster
                        if uncertainty_weights:
                            weights = 1 / contour_area[merging][merged[merging]]
                        else:
                            weights = area_top[merging][merged[merging]]

                        dr_avg = weighted_averaged_dr(
                            positions[merging, 0][merged[merging]],
                            positions[merging, 1][merged[merging]],
                            weights,
                        )
                        # do we really merge the peaklets?
                        dr_threshold_ = thresholds_interpolation(
                            np.log10(area_top[merging][merged[merging]].sum()),
                            dr_thresholds,
                        )
                        merge = dr_avg < dr_threshold_
                    else:
                        merge = True

                    if merge:
                        merged_peak = MergedS2s.merge_two_peaks(
                            peaks[direction[0]], peaks[direction[1]]
                        )
                        # check if the merged peak is too long
                        if merged_peak["endtime"] - merged_peak["time"] > max_duration:
                            raise ValueError("Merged S2 is too long")
                    else:
                        merged[index] = False
                        merged_peak = peaks[i]

                    # update merging peaks and boundaries
                    # update ALL peaks in the slice, this is only temporarily needed
                    # because we can not foresee in which order the peaks will be merged
                    peaks[merging] = merged_peak
                    start_index[merging] = start_index[merging.start]
                    end_index[merging] = end_index[merging.stop - 1]
                    left_bounds[merging] = left_bounds[merging.start]
                    right_bounds[merging] = right_bounds[merging.stop - 1]

                if diagnosing:
                    if left:
                        p_values.append(p[0])
                    else:
                        p_values.append(p[1])
                    if sparse_xy:
                        dr_avgs.append(dr_avg)
                    else:
                        dr_avgs.append(0.0)

        if np.any(np.diff(start_index) < 0) or np.any(np.diff(end_index) < 0):
            raise ValueError("Indices are not sorted!")
        n_peaklets = end_index - start_index
        need_merging = n_peaklets > 1
        start_index = np.unique(start_index[need_merging])
        end_index = np.unique(end_index[need_merging])
        if diagnosing:
            return start_index, end_index, merged, merged_area, p_values, dr_avgs
        return start_index, end_index, merged

    @staticmethod
    def merge_peaklets(
        peaklets,
        start_merge_at,
        end_merge_at,
        merged,
        max_buffer=int(1e5),
        max_unmerged=None,
        return_all_peaks=False,
    ):
        if return_all_peaks:
            # execute earlier to prevent peaklets from being overwritten
            # mark the peaklets can be merged by time-density but not
            # by position-density as type FAR_XYPOS_S2_TYPE
            _far_xypos = np.copy(peaklets[~merged])
            _far_xypos["type"] = FAR_XYPOS_S2_TYPE
        # mark the peaklets can be merged by time-density and position-density as type 2
        merged_s2s = strax.merge_peaks(
            peaklets,
            start_merge_at,
            end_merge_at,
            merged=merged,
            max_buffer=max_buffer,
        )
        # we do not do another round of classification for simplicity
        merged_s2s["type"] = 2

        # if the number of type 20 peaklets inside a peak is larger than the threshold
        # or the area of type 20 peaklets inside a peak is larger than the threshold
        # mark the peaklets as type WIDE_XYPOS_S2_TYPE
        if max_unmerged is not None:
            n_de = []
            area = []
            area_de = []
            for i in range(len(merged_s2s)):
                sl = slice(start_merge_at[i], end_merge_at[i])
                n_de.append(np.sum(~merged[sl]))
                area.append(peaklets["area"][sl][merged[sl]].sum())
                area_de.append(peaklets["area"][sl][~merged[sl]].sum())
            n_de = np.array(n_de)
            area = np.array(area)
            area_de = np.array(area_de)
            # if the number of type 20 peaklets inside a peak is larger than the threshold
            mask = n_de > max_unmerged[0]
            # if the area of type 20 peaklets inside a peak is larger than the (both) threshold(s)
            mask |= (area_de > max_unmerged[1]) & (area_de > area * max_unmerged[2])
            merged_s2s["type"] = np.where(mask, WIDE_XYPOS_S2_TYPE, merged_s2s["type"])

        if return_all_peaks:
            merged_s2s = strax.sort_by_time(np.concatenate([_far_xypos, merged_s2s]))
        return merged_s2s


@numba.njit(cache=True, nogil=True)
def get_p_value(
    y,
    x,
    ly,
    ry,
    lx,
    rx,
    sigma,
    rough_seg,
    sigma_seg,
    rough_mu_bins,
    poisson_max_mu,
    poisson_survival_ratio,
    normal_panel,
    factorial_panel,
    sigma_panel,
    maxexp,
):
    # better constraint the mu in boundaries of waveform
    # because the number of bins is not high enough comprising for computation efficiency
    mu = rough_mu(y, x, rough_mu_bins)

    t0 = y["time"]
    y_integrated = normal_cdf(((ry - t0) - mu[:, None]) / sigma, normal_panel)
    y_integrated -= normal_cdf(((ly - t0) - mu[:, None]) / sigma, normal_panel)
    x_integrated = normal_cdf(((rx - t0) - mu[:, None]) / sigma, normal_panel)
    x_integrated -= normal_cdf(((lx - t0) - mu[:, None]) / sigma, normal_panel)
    # keep numerical stability
    eps = 2.220446049250313e-16  # np.finfo(float).eps
    y_integrated += eps
    x_integrated += eps

    pmf = get_posterior(
        mu,
        sigma,
        y["area"],
        y["area_decile_from_midpoint"],
        y_integrated,
        y["time"],
        y["median_time"],
        x["time"],
        x["median_time"],
        rough_seg,
        maxexp,
    ).flatten()

    non_zero = pmf > 0
    if not np.any(non_zero):
        return 0.0
    # the mask non_zero here is also implicit flatten
    ps = p_values(
        sigma_panel[non_zero],
        y["area"],
        y_integrated.flatten()[non_zero],
        x["area"],
        x_integrated.flatten()[non_zero],
        rough_seg,
        sigma_seg,
        poisson_max_mu,
        poisson_survival_ratio,
        normal_panel,
        factorial_panel,
    )

    p = np.sum(ps * pmf[non_zero])
    if np.isnan(p):
        raise RuntimeError("p-value is NaN")
    return p


@numba.njit(cache=True, nogil=True)
def thresholds_interpolation(log_area, thresholds):
    """Return threshold for log_area of the merged S2 with linear interpolation given the points in
    thresholds.

    :param log_area: Log 10 area of the merged S2
    :param thresholds: tuple (n, 2) of fix points for interpolation.

    """
    if log_area < thresholds[0, 0]:
        return thresholds[1, 0]
    if log_area > thresholds[0, -1]:
        return thresholds[1, -1]
    return np.interp(log_area, thresholds[0], thresholds[1])


@numba.njit(cache=True, nogil=True)
def rough_mu(y, x, rough_mu_bins):
    """Get rough bins for mu as the integration space."""
    mu = np.linspace(
        min(y["time"], x["time"]) - y["time"],
        max(y["endtime"], x["endtime"]) - y["time"],
        rough_mu_bins,
    )
    return mu


@numba.njit(cache=True, nogil=True)
def poisson_pmf(k, mu, factorial_panel):
    """Probability mass function of Poisson distribution, numba decorated."""
    return mu**k / np.exp(mu) / factorial_panel[k]


@numba.njit(cache=True, nogil=True)
def normal_cdf(x, normal_panel):
    """Cumulative density function of standard normal distribution, numba decorated."""
    return np.interp(x, normal_panel[0], normal_panel[1])


@numba.njit(cache=True, nogil=True)
def get_posterior(
    mu,
    sigma,
    y_area,
    y_area_decile_from_midpoint,
    y_integrated,
    y_time,
    y_median_time,
    x_time,
    x_median_time,
    rough_seg,
    maxexp,
):
    """Get the posterior of (mu, sigma) based on the waveform y."""
    # production of normal PDF
    y_t = y_area_decile_from_midpoint + y_median_time
    y_t = (y_t[:-1] + y_t[1:]) / 2
    y_nse = y_area / rough_seg
    sigma2 = 2 * sigma**2
    log_sigma = np.log(sigma)
    # dt is median between median time of two peaklets
    dt = np.abs((y_time - x_time) + (y_median_time - x_median_time))
    # because the number of electrons are uniformly distributed in y_t
    log_pdf = -(((y_t - mu[:, None]) ** 2).sum(axis=1) * (y_nse / 10))[:, None] / sigma2
    # complete normal distribution PDF log_sigma * y_nse
    # add log prior dt**2 / 4 / sigma2 + 2 * log_sigma
    # account for the bounded sampling of y np.log(y_integrated) * y_nse
    log_pdf -= np.log(y_integrated) * y_nse + log_sigma * (y_nse + 2) + dt**2 / 4 / sigma2
    # to prevent numerical overflow
    log_pdf -= log_pdf.max()
    log_pdf += maxexp
    pdf = np.exp(log_pdf)
    pmf = pdf / pdf.sum()
    if np.any(np.isnan(pmf)):
        raise RuntimeError("pmf is NaN")
    return pmf


@numba.njit(cache=True, nogil=True)
def p_values(
    sigma,
    y_area,
    y_integrated,
    x_area,
    x_integrated,
    rough_seg,
    sigma_seg,
    poisson_max_mu,
    poisson_survival_ratio,
    normal_panel,
    factorial_panel,
):
    """The p-value if observe the x_area given the y_area, numba decorated."""
    # distribution of allowed area
    expected_nse = x_integrated * y_area / y_integrated / rough_seg
    ps = []
    for mu in expected_nse:
        # the number of electron follows Poisson distribution
        if mu > poisson_max_mu:
            ps.append(1 - normal_cdf((x_area / rough_seg - mu) / np.sqrt(mu), normal_panel))
        else:
            cdf = poisson_pmf(0, mu, factorial_panel)
            k = 1
            p = 0
            while cdf < 1 - poisson_survival_ratio:
                pmf = poisson_pmf(k, mu, factorial_panel)
                cdf += pmf
                p += pmf * (
                    1 - normal_cdf((x_area - k * rough_seg) / (k**0.5 * sigma_seg), normal_panel)
                )
                k += 1
            ps.append(p)
    ps = np.array(ps)
    return ps


@numba.njit(cache=True, nogil=True)
def weighted_averaged_dr(x, y, weights):
    """Weighted average deviation from weighted average (x, y)"""
    mask = weights > 0
    mask &= ~np.isnan(x)
    mask &= ~np.isnan(x)
    # do not merge any S2 looks weird
    if not np.all(mask):
        return np.nan
    x_avg = np.average(x[mask], weights=weights[mask])
    y_avg = np.average(y[mask], weights=weights[mask])
    dr = np.sqrt((x - x_avg) ** 2 + (y - y_avg) ** 2)
    dr_avg = np.average(dr[mask], weights=weights[mask])
    return dr_avg



import typing
import numpy as np
import strax

from straxen.plugins.aqmon_hits.aqmon_hits import AqmonChannels

export, __all__ = strax.exporter()


# ### Veto hardware ###:
# V1495 busy veto module:
# Generates a 25 ns NIM pulse whenever a veto begins and a 25 ns NIM signal when it ends.
# A new start signal can occur only after the previous busy instance ended.
# 1ms (1e6 ns) - minimum busy veto length, or until the board clears its memory

# DDC10 High Energy Veto:
# 10ms (1e7 ns) - fixed HE veto length in XENON1T DDC10,
# in XENONnT it will be calibrated based on the length of large S2 SE tails
# The start/stop signals for the HEV are generated by the V1495 board


@export
class VetoIntervals(strax.ExhaustPlugin):
    """Find pairs of veto start and veto stop signals and the veto.

    duration between them:
     - busy_* <= V1495 busy veto for tpc channels
     - busy_he_* <= V1495 busy veto for high energy tpc channels
     - hev_* <= DDC10 hardware high energy veto
     - straxen_deadtime <= special case of deadtime introduced by the
       DAQReader-plugin

    """

    __version__ = "1.1.1"
    depends_on = "aqmon_hits"
    provides = "veto_intervals"
    data_kind = "veto_intervals"

    def infer_dtype(self):
        dtype = [
            (("veto interval [ns]", "veto_interval"), np.int64),
            (("veto signal type", "veto_type"), np.str_("U30")),
        ]
        dtype += strax.time_fields
        return dtype

    def setup(self):
        self.veto_names = ["busy_", "busy_he_", "hev_"]
        self.channel_map = {aq_ch.name.lower(): int(aq_ch) for aq_ch in AqmonChannels}

    def compute(self, aqmon_hits, start, end):
        # Allocate a nice big buffer and throw away the part we don't need later
        result = np.zeros(len(aqmon_hits) * len(self.veto_names), self.dtype)
        vetos_seen = 0

        for veto_name in self.veto_names:
            veto_hits_start = channel_select(aqmon_hits, self.channel_map[veto_name + "start"])
            veto_hits_stop = channel_select(aqmon_hits, self.channel_map[veto_name + "stop"])

            veto_hits_start, veto_hits_stop = self.handle_starts_and_stops_outside_of_run(
                veto_hits_start=veto_hits_start,
                veto_hits_stop=veto_hits_stop,
                chunk_start=start,
                chunk_end=end,
                veto_name=veto_name,
            )
            n_vetos = len(veto_hits_start)

            result["time"][vetos_seen : vetos_seen + n_vetos] = veto_hits_start["time"]
            result["endtime"][vetos_seen : vetos_seen + n_vetos] = veto_hits_stop["time"]
            result["veto_type"][vetos_seen : vetos_seen + n_vetos] = veto_name + "veto"

            vetos_seen += n_vetos

        # Straxen deadtime is special, it's a start and stop with no data
        # but already an interval so easily used here
        artificial_deadtime = aqmon_hits[
            (aqmon_hits["channel"] == AqmonChannels.ARTIFICIAL_DEADTIME)
        ]
        n_artificial = len(artificial_deadtime)

        if n_artificial:
            result[vetos_seen:n_artificial]["time"] = artificial_deadtime["time"]
            result[vetos_seen:n_artificial]["endtime"] = strax.endtime(artificial_deadtime)
            result[vetos_seen:n_artificial]["veto_type"] = "straxen_deadtime_veto"
            vetos_seen += n_artificial

        result = result[:vetos_seen]
        result["veto_interval"] = result["endtime"] - result["time"]
        argsort = strax.stable_argsort(result["time"])
        result = result[argsort]
        return result

    def handle_starts_and_stops_outside_of_run(
        self,
        veto_hits_start: np.ndarray,
        veto_hits_stop: np.ndarray,
        chunk_start: int,
        chunk_end: int,
        veto_name: str,
    ) -> typing.Tuple[np.ndarray, np.ndarray]:
        """We might be missing one start or one stop at the end of the run, set it to the chunk
        endtime if this is the case."""
        # Just for traceback info that we declare this here
        extra_start = []
        extra_stop = []
        missing_a_final_stop = (
            len(veto_hits_start)
            and len(veto_hits_stop)
            and veto_hits_start[-1]["time"] > veto_hits_stop["time"][-1]
        )
        missing_a_final_stop = missing_a_final_stop or (
            len(veto_hits_start) and not len(veto_hits_stop)
        )
        if missing_a_final_stop:
            # There is one *start* of the //end// of the run -> the
            # **stop** is missing (because it's outside of the run),
            # let's add one **stop** at the //end// of this chunk
            extra_stop = self.fake_hit(chunk_end)
            veto_hits_stop = np.concatenate([veto_hits_stop, extra_stop])
        if len(veto_hits_stop) - len(veto_hits_start) == 1:
            # There is one *stop* of the //beginning// of the run
            # -> the **start** is missing (because it's from before
            # starting the run), # let's add one **start** at the
            # //beginning// of this chunk
            extra_start = self.fake_hit(chunk_start)
            veto_hits_start = np.concatenate([extra_start, veto_hits_start])

        something_is_wrong = len(veto_hits_start) != len(veto_hits_stop)

        message = (
            f"Got inconsistent number of {veto_name} starts "
            f"({len(veto_hits_start)}) / stops ({len(veto_hits_stop)})."
        )
        if len(extra_start):
            message += " Despite the fact that we inserted one extra start at the beginning of the run."  # noqa
        elif len(extra_stop):
            message += (
                " Despite the fact that we inserted one extra stop at the end of the run."  # noqa
            )
        if something_is_wrong:
            raise ValueError(message)

        if np.any(veto_hits_start["time"] > veto_hits_stop["time"]):
            raise ValueError("Found veto's starting before the previous stopped")

        return veto_hits_start, veto_hits_stop

    @staticmethod
    def fake_hit(start, dt=1, length=1):
        hit = np.zeros(1, strax.hit_dtype)
        hit["time"] = start
        hit["dt"] = dt
        hit["length"] = length
        return hit


# Don't use @numba since numba doesn't like masking arrays, use numpy
def channel_select(rr, ch):
    """Return data from start/stop veto channel in the acquisition monitor (AM)"""
    return rr[rr["channel"] == ch]



import numpy as np
import numba
import strax
import straxen


export, __all__ = strax.exporter()


@export
class PeakAmbience(strax.OverlapWindowPlugin):
    """Calculate Ambience of peaks. Features are the number of lonehits, small S0, S1, S2 in a time
    window before peaks, and the number of small S2 in circle near the S2 peak in a time window.

    References:
        * v0.0.7 reference: xenon:xenonnt:ac:prediction:shadow_ambience
        * v0.1.0 reference: xenon:xenonnt:analysis:redefine_n_competing

    """

    __version__ = "0.1.0"
    depends_on = ("peak_basics", "peak_positions", "lone_hits")
    provides = "peak_ambience"
    data_kind = "peaks"
    save_when = strax.SaveWhen.EXPLICIT

    gain_model = straxen.URLConfig(
        infer_type=False, help="PMT gain model. Specify as URL or explicit value"
    )

    ambience_time_window_backward = straxen.URLConfig(
        default=int(2e6), type=int, track=True, help="Search for ambience in this time window [ns]"
    )

    ambient_radius = straxen.URLConfig(
        default=6.7, type=float, track=True, help="Search for ambience in this radius [cm]"
    )

    ambience_area_parameters = straxen.URLConfig(
        default=(30, 30, 60),
        type=(list, tuple),
        track=True,
        help="The upper limit of S0, S1, S2 area to be counted",
    )

    ambience_exponents = straxen.URLConfig(
        default=(-1.0, -1.0),
        type=(list, tuple),
        help="The exponent of (delta t, delta r) when calculating ambience score",
    )

    def get_window_size(self):
        return (10 * self.ambience_time_window_backward, 0)

    @property
    def origin_dtype(self):
        return ["lh_before", "s0_before", "s1_before", "s2_before", "s2_near"]

    def infer_dtype(self):
        dtype = []
        for ambience in self.origin_dtype:
            dtype += [
                (
                    (f"Number of small {' '.join(ambience.split('_'))} a peak", f"n_{ambience}"),
                    np.int16,
                ),
                (
                    (f"Sum of small {' '.join(ambience.split('_'))} a peak", f"s_{ambience}"),
                    np.float32,
                ),
            ]
        dtype += [
            (
                ("Sum of small hits and peaks before a peak", "s_before"),
                np.float32,
            ),
        ]
        dtype += strax.time_fields
        return dtype

    def setup(self):
        self.to_pe = self.gain_model

    def compute(self, lone_hits, peaks):
        argsort = strax.stable_argsort(peaks["center_time"])
        _peaks = peaks[argsort].copy()
        result = np.zeros(len(peaks), self.dtype)
        _quick_assign(argsort, result, self.compute_ambience(lone_hits, peaks, _peaks))
        return result

    def compute_ambience(self, lone_hits, peaks, current_peak):
        # 1. Initialization
        result = np.zeros(len(current_peak), self.dtype)

        # 2. Define time window for each peak,
        # we will find small peaks & lone hits within these time windows
        roi = np.zeros(len(current_peak), dtype=strax.time_fields)
        roi["time"] = current_peak["center_time"] - self.ambience_time_window_backward
        roi["endtime"] = current_peak["center_time"]

        # 3. Calculate number and area sum of lonehits before a peak
        touching_windows = strax.touching_windows(lone_hits, roi)
        # Calculating ambience
        self.lonehits_ambience(
            current_peak,
            lone_hits,
            touching_windows,
            self.ambience_exponents,
            result["n_lh_before"],
            result["s_lh_before"],
            self.to_pe,
        )

        # 4. Calculate number and area sum of small S0, S1, S2 before a peak
        for stype, area in zip([0, 1, 2], self.ambience_area_parameters):
            mask_pre = (peaks["type"] == stype) & (peaks["area"] < area)
            touching_windows = strax.touching_windows(peaks[mask_pre], roi)
            # Calculating ambience
            self.peaks_ambience(
                current_peak,
                peaks[mask_pre],
                touching_windows,
                self.ambience_exponents,
                -1,
                result[f"n_s{stype}_before"],
                result[f"s_s{stype}_before"],
            )

        # 5. Calculate number and area sum of small S2 near(in (x,y) space) a S2 peak
        mask_pre = (peaks["type"] == 2) & (peaks["area"] < self.ambience_area_parameters[2])
        touching_windows = strax.touching_windows(peaks[mask_pre], roi)
        # Calculating ambience
        self.peaks_ambience(
            current_peak,
            peaks[mask_pre],
            touching_windows,
            self.ambience_exponents,
            self.ambient_radius,
            result["n_s2_near"],
            result["s_s2_near"],
        )

        # 6. Set time and endtime for peaks
        result["time"] = current_peak["time"]
        result["endtime"] = strax.endtime(current_peak)

        # 7. Calculate sum of small hits and peaks before a peak
        result["s_before"] = (
            result["s_lh_before"]
            + result["s_s0_before"]
            + result["s_s1_before"]
            + result["s_s2_before"]
        )
        return result

    @staticmethod
    @numba.njit
    def lonehits_ambience(
        peaks, pre_hits, touching_windows, exponents, num_array, sum_array, to_pe
    ):
        # Function to find lonehits before a peak
        # creating_hit is the lonehit creating ambience
        # suspicious_peak is the suspicious peak in the ambience created by creating_hit
        for p_i, suspicious_peak in enumerate(peaks):
            indices = touching_windows[p_i]
            for idx in range(indices[0], indices[1]):
                creating_hit = pre_hits[idx]
                dt = suspicious_peak["center_time"] - creating_hit["time"]
                if (dt <= 0) or (creating_hit["area"] <= 0):
                    continue
                num_array[p_i] += 1
                # Sometimes we may interested in sum of area / dt
                s = creating_hit["area"] * to_pe[creating_hit["channel"]]
                s *= dt ** exponents[0]
                sum_array[p_i] += s

    @staticmethod
    @numba.njit
    def peaks_ambience(
        peaks,
        pre_peaks,
        touching_windows,
        exponents,
        ambient_radius,
        num_array,
        sum_array,
    ):
        # Function to find S0, S1, S2 before or near a peak
        # creating_peak is the peak creating ambience
        # suspicious_peak is the suspicious peak in the ambience created by creating_peak
        for p_i, suspicious_peak in enumerate(peaks):
            indices = touching_windows[p_i]
            for idx in range(indices[0], indices[1]):
                creating_peak = pre_peaks[idx]
                r = distance_in_xy(suspicious_peak, creating_peak)
                dt = suspicious_peak["center_time"] - creating_peak["center_time"]
                if dt <= 0:
                    continue
                if (ambient_radius < 0) or (r <= ambient_radius):
                    num_array[p_i] += 1
                    # Sometimes we may interested in sum of area / dt
                    s = creating_peak["area"] * dt ** exponents[0]
                    # Sometimes we may interested in sum of area / r
                    if ambient_radius > 0:
                        s *= r ** exponents[1]
                    sum_array[p_i] += s


@numba.njit
def distance_in_xy(peak_a, peak_b):
    """Distance between S2s in (x,y)"""
    return np.sqrt((peak_a["x"] - peak_b["x"]) ** 2 + (peak_a["y"] - peak_b["y"]) ** 2)


@numba.njit
def _quick_assign(indices, results, inputs):
    for i, r in zip(indices, inputs):
        results[i] = r



from typing import Tuple

import numpy as np
import strax
import straxen
from straxen.common import rotate_perp_wires
from straxen.plugins.defaults import DEFAULT_POSREC_ALGO

export, __all__ = strax.exporter()


@export
class CorrectedAreas(strax.Plugin):
    """Plugin which applies light collection efficiency maps and electron life time to the data.

    Computes the cS1/cS2 for the main/alternative S1/S2 as well as the
    corrected life time.
    Note:
        Please be aware that for both, the main and alternative S1, the
        area is corrected according to the xy-position of the main S2.
        There are now 3 components of cS2s: cs2_top, cS2_bottom and cs2.
        cs2_top and cs2_bottom are corrected by the corresponding maps,
        and cs2 is the sum of the two.

    N-1 corrections are also provided, where each variable has all corrections
    applied except for one specific correction. This allows studying the impact
    of individual corrections.

    The following corrections are applied:
    - Peak reconstruction bias correction (corrects for bias in peak finding algorithm)
    - S1 xyz correction (light collection efficiency)
    - S2 xy correction (light collection efficiency)
    - Time-dependent light yield correction
    - Single electron gain (SEG) and extraction efficiency (EE) correction (partition,time)
    - Photoionization correction for S2 bottom
    - Electron lifetime correction

    """

    __version__ = "0.5.5"

    depends_on: Tuple[str, ...] = ("event_basics", "event_positions")

    # Descriptor configs
    elife = straxen.URLConfig(
        default="xedocs://electron_lifetimes?attr=value&run_id=plugin.run_id&version=ONLINE",
        help="electron lifetime in [ns]",
    )

    default_reconstruction_algorithm = straxen.URLConfig(
        default=DEFAULT_POSREC_ALGO, help="default reconstruction algorithm that provides (x,y)"
    )
    s1_xyz_map = straxen.URLConfig(
        default="xedocs://s1_xyz_maps"
        "?run_id=plugin.run_id"
        "&algorithm=plugin.default_reconstruction_algorithm&attr=map&version=ONLINE",
        cache=True,
    )
    s2_xy_map = straxen.URLConfig(
        default="xedocs://s2_xy_maps"
        "?run_id=plugin.run_id"
        "&algorithm=plugin.default_reconstruction_algorithm&attr=map&version=ONLINE",
        cache=True,
    )

    # average SE gain for a given time period. default to the value of this run in ONLINE model
    # thus, by default, there will be no time-dependent correction according to se gain
    avg_se_gain = straxen.URLConfig(
        default="xedocs://avg_se_gains?run_id=plugin.run_id&version=ONLINE&attr=value",
        help="Nominal single electron (SE) gain in PE / electron extracted. "
        "Data will be corrected to this value",
    )

    # se gain for this run, allowing for using xedocs. default to online
    se_gain = straxen.URLConfig(
        default="take://objects-to-dict://"
        "xedocs://se_gains"
        "?partition=all_tpc&run_id=plugin.run_id&sort=partition"
        "&as_list=True&key_attr=partition&value_attr=value&take=all_tpc&version=ONLINE",
        help="Actual SE gain for a given run (allows for time dependence)",
    )

    # relative extraction efficiency which can change with time and modeled by xedocs.
    rel_extraction_eff = straxen.URLConfig(
        default="take://objects-to-dict://"
        "xedocs://rel_extraction_effs"
        "?partition=all_tpc&run_id=plugin.run_id&sort=partition"
        "&as_list=True&key_attr=partition&value_attr=value&take=all_tpc&version=ONLINE",
        help="Relative extraction efficiency for this run (allows for time dependence)",
    )

    # relative light yield
    # defaults to no correction
    rel_light_yield = straxen.URLConfig(
        default="xedocs://relative_light_yield?attr=value&run_id=plugin.run_id&version=ONLINE",
        help="Relative light yield (allows for time dependence)",
    )

    # b parameter for z-dependent relative light yield correction
    b_rel_light_yield = straxen.URLConfig(
        default=0.0,
        help="b parameter for z-dependent relative light yield correction",
    )

    # slope parameter for z-dependent relative light yield correction
    slope_rel_light_yield = straxen.URLConfig(
        default=0.0,
        help="Slope parameter for z-dependent relative light yield correction",
    )

    # Single electron gain partition
    # AB and CD partitons distiguished based on
    # linear and circular regions
    # SR0 values set as default
    # https://xe1t-wiki.lngs.infn.it/doku.php?id=jlong:sr0_2_region_se_correction
    # https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:noahhood:corrections:se_gain_ee_final
    single_electron_gain_partition = straxen.URLConfig(
        default={"linear": 28, "circular": 60},
        help=(
            "Two distinct patterns of evolution of single electron corrections between AB and CD. "
            "Distinguish thanks to linear and circular regions"
        ),
    )

    # cS2 AFT correction due to photoionization
    # https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:zihao:sr1_s2aft_photonionization_correction
    cs2_bottom_top_ratio_correction = straxen.URLConfig(
        default=1, help="Scaling factor for cS2 AFT correction due to photoionization"
    )

    # S1 Peak Reconstruction Bias Map
    s1_bias_map = straxen.URLConfig(
        default="itp_map://resource://xedocs://peak_reconstruction_bias"
        "?attr=value&run_id=plugin.run_id&signal=s1&fmt=json&version=ONLINE",
        help="Interpolation map for S1 peak bias correction. "
        "Bias is defined as (reconstructed / raw) - 1."
        "So, the bias correction is reconstructed / (1 + bias).",
    )

    # S2 Peak Reconstruction Bias Map
    s2_bias_map = straxen.URLConfig(
        default="itp_map://resource://xedocs://peak_reconstruction_bias"
        "?attr=value&run_id=plugin.run_id&signal=s2&fmt=json&version=ONLINE",
        help="Interpolation map for S2 peak bias correction. "
        "Bias is defined as (reconstructed / raw) - 1."
        "So, the bias correction is reconstructed / (1 + bias).",
    )

    def infer_dtype(self):
        dtype = []
        dtype += strax.time_fields

        for peak_type, peak_name in zip(["", "alt_"], ["main", "alternate"]):
            # S1 corrections
            dtype += [
                (f"{peak_type}cs1", np.float32, f"Corrected area of {peak_name} S1 [PE]"),
                (
                    f"{peak_type}cs1_wo_timecorr",
                    np.float32,
                    f"Corrected area of {peak_name} S1 (without rel LY evolution correction) [PE]",
                ),
                (
                    f"{peak_type}cs1_wo_xyzcorr",
                    np.float32,
                    f"Corrected area of {peak_name} S1 (without xyz position correction) [PE]",
                ),
                (
                    f"{peak_type}cs1_wo_peakbiascorr",
                    np.float32,
                    f"Corrected area of {peak_name} S1 (without peak bias correction) [PE]",
                ),
            ]

            # Intermediate S2 corrections for studying correction order
            names = ["_wo_segee_picorr", "_before_pi", "_before_elife", ""]

            # Define what's "before" and "after" for each stage
            stage_definitions = {
                "_wo_segee_picorr": ("SEG/EE + photoionization", "peak bias + S2 xy + elife"),
                "_before_pi": ("photoionization + elife", "peak bias + S2 xy + SEG/EE"),
                "_before_elife": ("elife", "peak bias + S2 xy + SEG/EE + photoionization"),
                "": (None, None),
            }

            for name in names:
                before, after = stage_definitions[name]
                if before:
                    description = f" (before {before}, after {after})"
                else:
                    description = ""

                main_comment = f"Corrected area of {peak_name} S2{description} [PE]"
                aft_comment = (
                    f"Fraction of area seen by the top PMT array for corrected "
                    f"{peak_name} S2{description}"
                )

                dtype += [
                    (f"{peak_type}cs2{name}", np.float32, main_comment),
                    (f"{peak_type}cs2_area_fraction_top{name}", np.float32, aft_comment),
                ]

            # N-1 corrections for S2, list them in the order of application
            n1_versions = {
                "peakbiascorr": "peak bias",
                "xycorr": "xy position",
                "segee": "SEG/EE",
                "picorr": "photoionization",
                "elifecorr": "electron lifetime",
                "timecorr": "all time dependent",  # this is a N-M special case
            }
            for corr_name, corr_desc in n1_versions.items():
                main_comment = (
                    f"Corrected area of {peak_name} S2 (without {corr_desc} correction) [PE]"
                )
                aft_comment = (
                    f"Fraction of area seen by the top PMT array for corrected "
                    f"{peak_name} S2 (without {corr_desc} correction)"
                )
                dtype += [
                    (f"{peak_type}cs2_wo_{corr_name}", np.float32, main_comment),
                    (f"{peak_type}cs2_area_fraction_top_wo_{corr_name}", np.float32, aft_comment),
                ]
        return dtype

    def ab_region(self, x, y):
        new_x, new_y = rotate_perp_wires(x, y)
        cond = new_x < self.single_electron_gain_partition["linear"]
        cond &= new_x > -self.single_electron_gain_partition["linear"]
        cond &= new_x**2 + new_y**2 < self.single_electron_gain_partition["circular"] ** 2
        return cond

    def cd_region(self, x, y):
        return ~self.ab_region(x, y)

    def s2_map_names(self):
        # S2 top and bottom are corrected separately, and cS2 total is the sum of the two
        # figure out the map name
        if len(self.s2_xy_map.map_names) > 1:
            s2_top_map_name = "map_top"
            s2_bottom_map_name = "map_bottom"
        else:
            s2_top_map_name = "map"
            s2_bottom_map_name = "map"

        return s2_top_map_name, s2_bottom_map_name

    def seg_ee_correction_preparation(self):
        """Get single electron gain and extraction efficiency options."""
        self.regions = {"ab": self.ab_region, "cd": self.cd_region}

        # setup SEG and EE corrections
        # if they are dicts, we just leave them as is
        # if they are not, we assume they are floats and
        # create a dict with the same correction in each region
        if isinstance(self.se_gain, dict):
            seg = self.se_gain
        else:
            seg = {key: self.se_gain for key in self.regions}

        if isinstance(self.avg_se_gain, dict):
            avg_seg = self.avg_se_gain
        else:
            avg_seg = {key: self.avg_se_gain for key in self.regions}

        if isinstance(self.rel_extraction_eff, dict):
            ee = self.rel_extraction_eff
        else:
            ee = {key: self.rel_extraction_eff for key in self.regions}

        return seg, avg_seg, ee

    def rel_light_yield_correction(self, events):
        """Compute relative light yield correction (z- and t-dependent)."""

        a = self.slope_rel_light_yield * (self.rel_light_yield - 1)
        b = self.b_rel_light_yield

        # Compute full z- and t-dependent correction
        rel_ly_zt_corr = self.rel_light_yield * (a * (events["z"] ** 2 + b * events["z"]) + 1)

        return rel_ly_zt_corr

    def apply_s2_corrections(
        self,
        s2_area,
        s2_aft,
        s2_bias_correction,
        s2_xy_correction_top,
        s2_xy_correction_bottom,
        seg_ee_corr,
        pi_corr_bottom,
        elife_correction,
    ):
        """Apply S2 corrections and return various corrected areas.

        Returns:
            cs2,
            cs2_area_fraction_top,
            cs2_before_pi,
            cs2_area_fraction_top_before_pi,
            cs2_before_elife,
            cs2_area_fraction_top_before_elife,
            cs2_wo_segee_picorr,
            cs2_area_fraction_top_wo_segee_picorr,

        """
        # Base areas
        s2_area_top = s2_area * s2_aft
        s2_area_bottom = s2_area * (1 - s2_aft)

        # Fully corrected S2 (in stages for intermediate variables)
        cs2_top_before_pi = s2_area_top / s2_bias_correction / s2_xy_correction_top / seg_ee_corr
        cs2_bottom_before_pi = (
            s2_area_bottom / s2_bias_correction / s2_xy_correction_bottom / seg_ee_corr
        )
        cs2_before_pi = cs2_top_before_pi + cs2_bottom_before_pi

        cs2_top_before_elife = cs2_top_before_pi
        cs2_bottom_before_elife = cs2_bottom_before_pi * pi_corr_bottom
        cs2_before_elife = cs2_top_before_elife + cs2_bottom_before_elife

        cs2 = cs2_before_elife * elife_correction
        cs2_top_final = cs2_top_before_elife * elife_correction

        cs2_top_wo_segee_picorr = s2_area_top / s2_bias_correction / s2_xy_correction_top
        cs2_bottom_wo_segee_picorr = s2_area_bottom / s2_bias_correction / s2_xy_correction_bottom
        cs2_wo_segee_picorr = (
            cs2_top_wo_segee_picorr + cs2_bottom_wo_segee_picorr
        ) * elife_correction
        cs2_top_wo_segee_picorr = cs2_top_wo_segee_picorr * elife_correction
        return (
            cs2,
            cs2_top_final / cs2,
            cs2_before_pi,
            cs2_top_before_pi / cs2_before_pi,
            cs2_before_elife,
            cs2_top_before_elife / cs2_before_elife,
            cs2_wo_segee_picorr,
            cs2_top_wo_segee_picorr / cs2_wo_segee_picorr,
        )

    def compute(self, events):
        result = np.zeros(len(events), self.dtype)
        result["time"] = events["time"]
        result["endtime"] = events["endtime"]

        # S1 corrections depend on the actual corrected event position.
        # We use this also for the alternate S1; for e.g. Kr this is
        # fine as the S1 correction varies slowly.
        event_positions = np.vstack([events["x"], events["y"], events["z"]]).T

        # S1 corrections
        for peak_type in ["", "alt_"]:
            # Correction factors
            s1_area = events[f"{peak_type}s1_area"]
            s1_bias_correction = 1 + self.s1_bias_map(s1_area.reshape(-1, 1)).flatten()
            s1_xyz_correction = self.s1_xyz_map(event_positions)
            s1_time_correction = self.rel_light_yield_correction(events)

            # Apply all corrections
            result[f"{peak_type}cs1"] = (
                s1_area / s1_bias_correction / s1_xyz_correction / s1_time_correction
            )

            # N-1 corrections for S1
            result[f"{peak_type}cs1_wo_peakbiascorr"] = (
                s1_area / s1_xyz_correction / s1_time_correction
            )
            result[f"{peak_type}cs1_wo_xyzcorr"] = s1_area / s1_bias_correction / s1_time_correction
            result[f"{peak_type}cs1_wo_timecorr"] = s1_area / s1_bias_correction / s1_xyz_correction

        # S2 corrections
        s2_top_map_name, s2_bottom_map_name = self.s2_map_names()
        seg, avg_seg, ee = self.seg_ee_correction_preparation()

        for peak_type in ["", "alt_"]:
            s2_area = events[f"{peak_type}s2_area"]
            s2_aft = events[f"{peak_type}s2_area_fraction_top"]
            s2_positions = np.vstack([events[f"{peak_type}s2_x"], events[f"{peak_type}s2_y"]]).T

            # Correction factors, list them in the order of application:
            # 1. Peak bias correction
            # 2. S2 xy position correction
            # 3. SEG/EE correction
            # 4. Photoionization correction for S2 bottom
            # 5. Electron lifetime correction

            s2_bias_correction = 1 + self.s2_bias_map(s2_area.reshape(-1, 1)).flatten()
            s2_xy_correction_top = self.s2_xy_map(s2_positions, map_name=s2_top_map_name)
            s2_xy_correction_bottom = self.s2_xy_map(s2_positions, map_name=s2_bottom_map_name)

            seg_ee_corr = np.zeros(len(events))
            for partition, func in self.regions.items():
                mask = func(events[f"{peak_type}s2_x"], events[f"{peak_type}s2_y"])
                seg_ee_corr[mask] = seg[partition] / avg_seg[partition] * ee[partition]

            pi_corr_bottom = self.cs2_bottom_top_ratio_correction

            el_string = peak_type + "s2_interaction_" if peak_type == "alt_" else peak_type
            elife_correction = np.exp(events[f"{el_string}drift_time"] / self.elife)

            (
                result[f"{peak_type}cs2"],
                result[f"{peak_type}cs2_area_fraction_top"],
                result[f"{peak_type}cs2_before_pi"],
                result[f"{peak_type}cs2_area_fraction_top_before_pi"],
                result[f"{peak_type}cs2_before_elife"],
                result[f"{peak_type}cs2_area_fraction_top_before_elife"],
                result[f"{peak_type}cs2_wo_segee_picorr"],
                result[f"{peak_type}cs2_area_fraction_top_wo_segee_picorr"],
            ) = self.apply_s2_corrections(
                s2_area,
                s2_aft,
                s2_bias_correction,
                s2_xy_correction_top,
                s2_xy_correction_bottom,
                seg_ee_corr,
                pi_corr_bottom,
                elife_correction,
            )

            # N-1 corrections for S2
            # N-1: without peak bias
            (
                result[f"{peak_type}cs2_wo_peakbiascorr"],
                result[f"{peak_type}cs2_area_fraction_top_wo_peakbiascorr"],
            ) = self.apply_s2_corrections(
                s2_area,
                s2_aft,
                1,
                s2_xy_correction_top,
                s2_xy_correction_bottom,
                seg_ee_corr,
                pi_corr_bottom,
                elife_correction,
            )[
                :2
            ]

            # N-1: without xy position
            (
                result[f"{peak_type}cs2_wo_xycorr"],
                result[f"{peak_type}cs2_area_fraction_top_wo_xycorr"],
            ) = self.apply_s2_corrections(
                s2_area,
                s2_aft,
                s2_bias_correction,
                1,
                1,
                seg_ee_corr,
                pi_corr_bottom,
                elife_correction,
            )[
                :2
            ]

            # N-1: without SEG/EE
            (
                result[f"{peak_type}cs2_wo_segee"],
                result[f"{peak_type}cs2_area_fraction_top_wo_segee"],
            ) = self.apply_s2_corrections(
                s2_area,
                s2_aft,
                s2_bias_correction,
                s2_xy_correction_top,
                s2_xy_correction_bottom,
                1,
                pi_corr_bottom,
                elife_correction,
            )[
                :2
            ]

            # N-1: without photoionization
            (
                result[f"{peak_type}cs2_wo_picorr"],
                result[f"{peak_type}cs2_area_fraction_top_wo_picorr"],
            ) = self.apply_s2_corrections(
                s2_area,
                s2_aft,
                s2_bias_correction,
                s2_xy_correction_top,
                s2_xy_correction_bottom,
                seg_ee_corr,
                1,
                elife_correction,
            )[
                :2
            ]

            # N-1: without electron lifetime
            (
                result[f"{peak_type}cs2_wo_elifecorr"],
                result[f"{peak_type}cs2_area_fraction_top_wo_elifecorr"],
            ) = self.apply_s2_corrections(
                s2_area,
                s2_aft,
                s2_bias_correction,
                s2_xy_correction_top,
                s2_xy_correction_bottom,
                seg_ee_corr,
                pi_corr_bottom,
                1,
            )[
                :2
            ]

            # N-1: without any time dependent correction
            (
                result[f"{peak_type}cs2_wo_timecorr"],
                result[f"{peak_type}cs2_area_fraction_top_wo_timecorr"],
            ) = self.apply_s2_corrections(
                s2_area,
                s2_aft,
                s2_bias_correction,
                s2_xy_correction_top,
                s2_xy_correction_bottom,
                1,
                1,
                1,
            )[
                :2
            ]

        return result



"""Process a single run with straxen."""

import argparse
import datetime
import logging
import time
import os
import psutil
import json
import importlib
import sys


def parse_args():
    parser = argparse.ArgumentParser(
        description="Process a single run with straxen",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "run_id", metavar="RUN_ID", type=str, help="ID of the run to process; usually the run name."
    )
    parser.add_argument(
        "--package", default="straxen", help="Where to load the context from (straxen/cutax/pema)"
    )
    parser.add_argument("--context", default="xenonnt_online", help="Name of context to use")
    parser.add_argument(
        "--target",
        default="event_info",
        nargs="*",
        help="Target final data type to produce. Can be a list for multicore mode.",
    )
    parser.add_argument(
        "--context_kwargs", type=json.loads, help="Use a json-file to load the context with"
    )
    parser.add_argument(
        "--register_from_file", type=str, help="do st.register_all from a specified file"
    )
    parser.add_argument(
        "--config_kwargs", type=json.loads, help="Use a json-file to set the context to"
    )
    parser.add_argument(
        "--from_scratch",
        action="store_true",
        help=(
            "Start processing at raw_records, regardless of what data is available. "
            "Saving will ONLY occur to ./strax_data! If you already have the target"
            "data in ./strax_data, you need to delete it there first."
        ),
    )
    parser.add_argument(
        "--max_messages",
        default=4,
        type=int,
        help=(
            "Size of strax's internal mailbox buffers. "
            "Lower to reduce memory usage, at increasing risk of deadlocks."
        ),
    )
    parser.add_argument(
        "--timeout", default=None, type=int, help="Strax' internal mailbox timeout in seconds"
    )
    parser.add_argument(
        "--workers",
        default=1,
        type=int,
        help=(
            "Number of worker threads/processes. "
            "Strax will multithread (1/plugin) even if you set this to 1."
        ),
    )
    parser.add_argument(
        "--notlazy",
        action="store_true",
        help="Forbid lazy single-core processing. Not recommended.",
    )
    parser.add_argument("--multiprocess", action="store_true", help="Allow multiprocessing.")
    parser.add_argument(
        "--multi_target",
        action="store_true",
        help=(
            "Allow st.make to be called with multiple targets at once "
            "(otherwise loop over the target list)"
        ),
    )
    parser.add_argument(
        "--shm", action="store_true", help="Allow passing data via /dev/shm when multiprocessing."
    )
    parser.add_argument(
        "--profile_to",
        default="",
        help="Filename to output profile information to. If omitted,no profiling will occur.",
    )
    parser.add_argument(
        "--profile_ram",
        action="store_true",
        help=(
            "Use memory_profiler for a more accurate measurement of the "
            "peak RAM usage of the process."
        ),
    )
    parser.add_argument(
        "--diagnose_sorting",
        action="store_true",
        help="Diagnose sorting problems during processing",
    )
    parser.add_argument("--debug", action="store_true", help="Enable debug logging to stdout")
    parser.add_argument(
        "--build_lowlevel",
        action="store_true",
        help="Build low-level data even if the context forbids it.",
    )
    parser.add_argument(
        "--only_strax_data", action="store_true", help="Only use ./strax_data (if not on dali)."
    )
    parser.add_argument("--add_folder", default="", help="Also add folder to st.storage")
    parser.add_argument(
        "--print_alive",
        default=300,
        help="Print that straxer is still running every this many [seconds]",
    )
    return parser.parse_args()


def setup_context(args):
    # reimport to be safe
    import strax
    import straxen

    context_module = importlib.import_module(f"{args.package}.contexts")
    st = getattr(context_module, args.context)()

    if args.context_kwargs:
        logging.info(f"set context kwargs {args.context_kwargs}")
        st = getattr(context_module, args.context)(**args.context_kwargs)

    if args.config_kwargs:
        logging.info(f"set context options to {args.config_kwargs}")
        st.set_config(to_dict_tuple(args.config_kwargs))

    if args.register_from_file:
        register_to_context(st, args.register_from_file)

    if args.diagnose_sorting:
        st.set_config(dict(diagnose_sorting=True))

    st.context_config["allow_multiprocess"] = args.multiprocess
    st.context_config["allow_shm"] = args.shm
    st.context_config["allow_lazy"] = not (args.notlazy is True)

    if args.timeout is not None:
        st.context_config["timeout"] = args.timeout
    st.context_config["max_messages"] = args.max_messages

    if args.build_lowlevel:
        st.context_config["forbid_creation_of"] = tuple()
    else:
        st.context_config["forbid_creation_of"] = straxen.DAQReader.provides

    if args.from_scratch:
        for q in st.storage:
            q.take_only = ("raw_records",)
        st.storage.append(
            strax.DataDirectory("./strax_data", overwrite="always", provide_run_metadata=False)
        )
    if args.only_strax_data:
        for sf in st.storage:
            # Set all others to read only
            sf.readonly = True
        for sf in st.storage:
            if hasattr(sf, "path"):
                if sf.path == "./strax_data":
                    break
        else:
            st.storage += [strax.DataDirectory("./strax_data")]

    if args.add_folder != "":
        for sf in st.storage:
            # Set all others to read only
            sf.readonly = True
        if os.path.exists(args.add_folder):
            st.storage += [strax.DataDirectory(args.add_folder)]

    if st.is_stored(args.run_id, args.target):
        logging.warning("This data is already available. Straxer is done")
        sys.exit(0)
    return st


def run(args):
    logging.basicConfig(
        level=logging.DEBUG if args.debug else logging.INFO,
        format="%(asctime)s - %(threadName)s - %(name)s - %(levelname)s - %(message)s",
    )

    logging.info(f"Starting processing of run {args.run_id} until {args.target}")

    # These imports take a bit longer, so it's nicer
    # to do them after argparsing (so --help is fast)
    import strax
    import straxen
    import pandas as pd

    # For showing data availability below
    pd.options.display.max_rows = 999
    logging.info(
        straxen.print_versions(
            tuple({"strax", "straxen", args.package}), return_string=True, print_output=False
        ),
    )

    logging.info("Starting context")
    st = setup_context(args)
    logging.info("Context started")

    # Reactivate after https://github.com/XENONnT/straxen/issues/586
    logging.info(f"Checking availabilty")
    logging.info(f"Available\n{str(st.available_for_run(args.run_id))}")

    logging.info("Infer start/end")
    try:
        md = st.run_metadata(args.run_id)
        t_start = md["start"].replace(tzinfo=datetime.timezone.utc).timestamp()
        t_end = md["end"].replace(tzinfo=datetime.timezone.utc).timestamp()
        st.config["run_start_time"] = md["start"].timestamp()
        st.context_config["free_options"] = tuple(
            list(st.context_config["free_options"]) + ["run_start_time"]
        )
    except strax.RunMetadataNotAvailable:
        for t in ("raw_records", "records", "peaklets"):
            if st.is_stored(args.run_id, t):
                break
        t_start, t_end = st.estimate_run_start_and_end(args.run_id, t)
        t_start, t_end = t_start / 1e9, t_end / 1e9
    run_duration = t_end - t_start
    logging.info(f"Infer start/end: run is {run_duration:.1f} s")

    start_alive_thread(logging, args.print_alive)

    process = psutil.Process(os.getpid())
    peak_ram = 0

    def run_make(targets):
        """For target or a list of targets, run st.make and print the progress."""
        nonlocal peak_ram
        if not args.multi_target:
            logging.info(f"Checking if {targets} is stored")
            if st.is_stored(args.run_id, targets):
                logging.info(f"{args.run_id}:{targets} is stored")
                return
        logging.info(f"Start processing {args.run_id}:{targets}")

        def get_results():
            kwargs = dict(
                run_id=args.run_id,
                targets=targets,
                max_workers=int(args.workers),
                allow_multiple=args.multi_target,
                progress_bar=False,
                save=strax.to_str_tuple(targets),
            )

            if args.profile_to:
                with strax.profile_threaded(args.profile_to):
                    yield from st.get_iter(**kwargs)
            else:
                yield from st.get_iter(**kwargs)

        clock_start = None
        for i, d in enumerate(get_results()):
            mem_mb = process.memory_info().rss / 1e6
            peak_ram = max(mem_mb, peak_ram)

            if not len(d):
                logging.info(f"Got chunk {i}, but it is empty! Using {mem_mb:.1f} MB RAM.")
                continue

            # Compute detector/data time left
            time_end = d.end / 1e9
            dt = time_end - t_start
            time_left = t_end - time_end

            msg = (
                f"Got {len(d)} items. "
                f"Now {dt:.1f} sec / {100 * dt / run_duration:.1f}% into the run. "
                f"Using {mem_mb:.1f} MB RAM. "
            )
            if clock_start is not None:
                # Compute processing job clock time left
                d_clock = time.time() - clock_start
                clock_time_left = time_left / (dt / d_clock)
                msg += f"ETA {clock_time_left:.2f} sec."
            else:
                clock_start = time.time()
            logging.info(msg)
        logging.info(f"{targets} finished! Took {time.time() - clock_start:.1f} s")

    proc_start = time.time()
    if args.multi_target:
        run_make(args.target)
    else:
        for target in strax.to_str_tuple(args.target):
            run_make(target)
    logging.info(
        f"Straxer is done in {time.time() - proc_start :.1f} s! peak RAM usage was"
        f" ~{peak_ram:.1f} MB."
    )


def register_to_context(st, module: str):
    if not os.path.exists(module):
        raise FileNotFoundError(f"No such file {module}")
    assert module.endswith(".py"), "only py files please!"
    folder, file = os.path.split(module)
    sys.path.append(folder)
    to_register = importlib.import_module(os.path.splitext(file)[0])
    st.register_all(to_register)
    logging.info(f"Successfully registered {file}. Printing plugins")

    for key, plugin in st._plugin_class_registry.items():
        logging.info(f"{key}\t{plugin}")


def to_dict_tuple(res: dict):
    """Convert list configs to tuple configs."""
    res = res.copy()
    for k, v in res.copy().items():
        if isinstance(v, list):
            # Remove lists to tuples
            res[k] = tuple(_v if not isinstance(_v, list) else tuple(_v) for _v in v)
    return res


def print_is_alive(log, print_timeout):
    while True:
        log.info("Straxer still running")
        time.sleep(print_timeout)


def start_alive_thread(log, print_timeout):
    from threading import Thread

    thread = Thread(
        name="Ping alive", target=print_is_alive, args=(log, print_timeout), daemon=True
    )
    log.info(f"Starting thread to ping that we are still running")
    thread.start()


def main():
    args = parse_args()
    if args.profile_ram:
        from memory_profiler import memory_usage

        mem = memory_usage(proc=(run, (args,)))
        print(f"Memory profiler says peak RAM usage was: {max(mem):.1f} MB")
        sys.exit()
    else:
        sys.exit(run(args))


if __name__ == "__main__":
    main()



import strax
import numpy as np
from straxen.plugins.peaklets.peaklet_classification_som import som_additional_fields
from straxen.plugins.peaks.peaks_vanilla import PeaksVanilla

export, __all__ = strax.exporter()


@export
class PeaksSOM(PeaksVanilla):
    """Same as Peaks but include in addition SOM type field to be propagated to event_basics.

    Thus, only change dtype.

    """

    __version__ = "0.0.1"
    child_plugin = True

    def compute(self, peaklets, merged_s2s):
        som_additional = np.zeros(
            len(merged_s2s), dtype=strax.to_numpy_dtype(som_additional_fields)
        )
        strax.set_nan_defaults(som_additional)
        # make sure _merged_s2s and peaklets have same dtype
        _merged_s2s = strax.merge_arrs([merged_s2s, som_additional], dtype=peaklets.dtype)
        peaks = super().compute(peaklets, _merged_s2s)
        return peaks



import numpy as np
import numba
import strax
import straxen

from .peak_ambience import _quick_assign
from ..events import Events

export, __all__ = strax.exporter()


@export
class PeakNearestTriggering(Events):
    """Time difference and properties of the nearest triggering peaks in the left and right
    direction of peaks."""

    __version__ = "0.0.0"
    depends_on = ("peak_basics", "peak_proximity")
    provides = "peak_nearest_triggering"
    data_kind = "peaks"
    save_when = strax.SaveWhen.EXPLICIT

    shadow_time_window_backward = straxen.URLConfig(
        default=int(1e9),
        type=int,
        track=True,
        help="Search for peaks casting time & position shadow in this time window [ns]",
    )

    only_trigger_min_area = straxen.URLConfig(
        default=False,
        type=bool,
        track=True,
        help="Whether only require the triggering peak to have area larger than trigger_min_area",
    )

    def infer_dtype(self):
        dtype = []
        common_descr = "of the nearest triggering peak on the"
        for direction in ["left", "right"]:
            dtype += [
                (
                    (f"time difference {common_descr} {direction} [ns]", f"{direction}_dtime"),
                    np.int64,
                ),
                ((f"time {common_descr} {direction} [ns]", f"{direction}_time"), np.int64),
                ((f"endtime {common_descr} {direction} [ns]", f"{direction}_endtime"), np.int64),
                (
                    (f"center_time {common_descr} {direction} [ns]", f"{direction}_center_time"),
                    np.int64,
                ),
                ((f"type {common_descr} {direction}", f"{direction}_type"), np.int8),
                (
                    (f"proximity_score {common_descr} {direction}", f"{direction}_proximity_score"),
                    np.float32,
                ),
                (
                    (
                        f"n_competing_left {common_descr} {direction}",
                        f"{direction}_n_competing_left",
                    ),
                    np.int32,
                ),
                ((f"n_competing {common_descr} {direction}", f"{direction}_n_competing"), np.int32),
                ((f"area {common_descr} {direction} [PE]", f"{direction}_area"), np.float32),
            ]
        dtype += strax.time_fields
        return dtype

    def get_window_size(self):
        # This method is required by the OverlapWindowPlugin class
        return 10 * self.shadow_time_window_backward

    def compute(self, peaks):
        argsort = strax.stable_argsort(peaks["center_time"])
        _peaks = peaks[argsort].copy()
        result = np.zeros(len(peaks), self.dtype)
        _quick_assign(argsort, result, self.compute_triggering(peaks, _peaks))
        return result

    def compute_triggering(self, peaks, current_peak):
        # sort peaks by center_time,
        # because later we will use center_time to find the nearest peak
        _peaks = strax.stable_sort(peaks, order="center_time")
        # only looking at triggering peaks
        if self.only_trigger_min_area:
            _is_triggering = _peaks["area"] > self.trigger_min_area
        else:
            _is_triggering = self._is_triggering(_peaks)
        _peaks = _peaks[_is_triggering]
        # init result
        result = np.zeros(len(current_peak), self.dtype)
        strax.set_nan_defaults(result)

        # use center_time as the anchor of things
        things = np.zeros(len(_peaks), dtype=strax.time_fields)
        things["time"] = _peaks["center_time"]
        things["endtime"] = _peaks["center_time"]
        # also se center_time as the anchor of containers
        containers = np.zeros(len(current_peak), dtype=strax.time_fields)
        containers["time"] = current_peak["center_time"] - self.shadow_time_window_backward
        containers["endtime"] = current_peak["center_time"] + self.shadow_time_window_backward

        # find indices of the nearest peaks in left and right direction
        split_peaks = strax.touching_windows(things, containers)
        left_indices, right_indices = self.nearest_indices(
            current_peak["center_time"],
            _peaks["center_time"],
            split_peaks,
        )
        # assign fields
        result["left_dtime"] = np.where(
            left_indices != -1,
            current_peak["center_time"] - _peaks["center_time"][left_indices],
            self.shadow_time_window_backward,
        )
        result["right_dtime"] = np.where(
            right_indices != -1,
            _peaks["center_time"][right_indices] - current_peak["center_time"],
            self.shadow_time_window_backward,
        )
        for field in [
            "time",
            "endtime",
            "center_time",
            "type",
            "proximity_score",
            "n_competing_left",
            "n_competing",
            "area",
        ]:
            result["left_" + field] = np.where(
                left_indices != -1, _peaks[field][left_indices], result["left_" + field]
            )
            result["right_" + field] = np.where(
                right_indices != -1, _peaks[field][right_indices], result["right_" + field]
            )
        result["time"] = current_peak["time"]
        result["endtime"] = strax.endtime(current_peak)
        for direction in ["left", "right"]:
            assert np.all(result[f"{direction}_dtime"] > 0), f"{direction}_dtime should be positive"
        return result

    @staticmethod
    @numba.njit
    def nearest_indices(reference_time, nearest_time, touching_windows):
        """Find the nearest indices in the left and right direction."""
        left_indices = np.full(len(reference_time), -1, dtype=np.int64)
        right_indices = np.full(len(reference_time), -1, dtype=np.int64)
        for r_i, r in enumerate(reference_time):
            indices = touching_windows[r_i]
            if indices[0] == indices[1]:
                continue
            center_time = nearest_time[indices[0] : indices[1]]
            left_dtime = r - center_time[0]
            right_dtime = center_time[-1] - r
            # make sure that the dtime is positive
            for p_i, t in enumerate(center_time):
                if t < r and r - t <= left_dtime:
                    left_dtime = r - t
                    left_indices[r_i] = indices[0] + p_i
                elif t > r and t - r <= right_dtime:
                    right_dtime = t - r
                    right_indices[r_i] = indices[0] + p_i
        return left_indices, right_indices



import numpy as np
import strax
from straxen.plugins.defaults import HE_PREAMBLE
from straxen.plugins.peaks.peaks_vanilla import PeaksVanilla

export, __all__ = strax.exporter()


@export
class PeaksHighEnergy(PeaksVanilla):
    __doc__ = HE_PREAMBLE + (PeaksVanilla.__doc__ or "")

    __version__ = "0.0.1"
    depends_on = ("peaklets_he", "peaklet_classification_he", "merged_s2s_he")
    data_kind = "peaks_he"
    provides = "peaks_he"
    child_ends_with = "_he"

    def infer_dtype(self):
        return strax.merged_dtype(
            (
                self.deps["peaklets_he"].dtype_for("peaklets"),
                self.deps["merged_s2s_he"].indicator_dtype,
            )
        )

    def compute(self, peaklets_he, merged_s2s_he):
        indicator_dtype = self.deps["merged_s2s_he"].indicator_dtype
        _peaklets_he = strax.merge_arrs(
            [peaklets_he, np.zeros(len(peaklets_he), dtype=indicator_dtype)],
            dtype=strax.merged_dtype((peaklets_he.dtype, indicator_dtype)),
        )
        return super().compute(_peaklets_he, merged_s2s_he)



import numpy as np
import straxen
import numba
import typing as ty

import strax

export, __all__ = strax.exporter()


@export
class nVETOEventPositions(strax.Plugin):
    """Plugin which computes the interaction position in the nveto as an azimuthal angle."""

    __version__ = "0.1.1"

    depends_on = ("events_nv", "hitlets_nv")
    data_kind = "events_nv"
    provides = "event_positions_nv"
    compressor = "zstd"

    position_max_time_nv = straxen.URLConfig(
        default=20,
        infer_type=False,
        help="Time [ns] within an event use to compute the azimuthal angle of the event.",
    )

    nveto_pmt_position_map = straxen.URLConfig(
        help="nVeto PMT position mapfile",
        default="resource://nveto_pmt_position.csv?fmt=csv",
        infer_type=False,
    )

    def infer_dtype(self):
        return veto_event_positions_dtype()

    def setup(self):
        npmt_pos = self.nveto_pmt_position_map
        # Use records instead of a dataframe.
        self.pmt_properties = npmt_pos.to_records(index=False)

    def compute(self, events_nv, hitlets_nv):
        event_angles = np.zeros(len(events_nv), dtype=self.dtype)

        # Split hitlets by containment, works since we updated event start/end in
        # compute_event_properties.
        hits_in_events = strax.split_by_containment(hitlets_nv, events_nv)

        # Compute hitlets within the first x ns of event:
        hits_in_events, n_prompt = first_hitlets(hits_in_events, self.position_max_time_nv)
        event_angles["n_prompt_hitlets"] = n_prompt

        # Compute azimuthal angle and xyz positions:
        angle = get_average_angle(hits_in_events, self.pmt_properties)
        event_angles["angle"] = angle
        compute_positions(event_angles, hits_in_events, self.pmt_properties)
        strax.copy_to_buffer(events_nv, event_angles, "_copy_events_nv")

        return event_angles


def veto_event_positions_dtype() -> list:
    dtype = []
    dtype += strax.time_fields
    dtype += [
        (
            (
                'Number of prompt hitlets within the first "position_max_time_nv" ns of the event.',
                "n_prompt_hitlets",
            ),
            np.int16,
        ),
        (
            ("Azimuthal angle, where the neutron capture was detected in [0, 2 pi).", "angle"),
            np.float32,
        ),
        (("Area weighted mean of position in x [mm]", "pos_x"), np.float32),
        (("Area weighted mean of position in y [mm]", "pos_y"), np.float32),
        (("Area weighted mean of position in z [mm]", "pos_z"), np.float32),
        (("Weighted variance of position in x [mm]", "pos_x_spread"), np.float32),
        (("Weighted variance of position in y [mm]", "pos_y_spread"), np.float32),
        (("Weighted variance of position in z [mm]", "pos_z_spread"), np.float32),
    ]
    return dtype


@numba.njit(cache=True, nogil=True)
def compute_positions(
    event_angles: np.ndarray,
    contained_hitlets: numba.typed.typedlist.List,
    pmt_pos: np.ndarray,
    start_channel: int = 2000,
):
    """Function which computes some artificial event position for a given neutron/muon-veto event.
    The position is computed based on a simple area weighted mean. Please note that the event
    position can be reconstructed in unphysical regions like being within the TPC.

    :param event_angles: Result array of the veto_event_position dtype. The result is updated
        inplace.
    :param contained_hitlets: Hitlets contained in each event.
    :param pmt_pos: Position of the veto PMTs
    :param start_channel: Starting channel of the detector.

    """
    for e_angles, hitlets in zip(event_angles, contained_hitlets):
        prompt_event_area = np.sum(hitlets["area"])
        if prompt_event_area:
            ch = hitlets["channel"] - start_channel
            pos_x = pmt_pos["x"][ch]
            pos_y = pmt_pos["y"][ch]
            pos_z = pmt_pos["z"][ch]

            e_angles["pos_x"] = np.sum(pos_x * hitlets["area"]) / prompt_event_area
            e_angles["pos_y"] = np.sum(pos_y * hitlets["area"]) / prompt_event_area
            e_angles["pos_z"] = np.sum(pos_z * hitlets["area"]) / prompt_event_area
            w = hitlets["area"] / prompt_event_area  # normalized weights
            if len(hitlets) and np.sum(w) > 0:
                e_angles["pos_x_spread"] = np.sqrt(
                    np.sum(w * (pos_x - e_angles["pos_x"]) ** 2) / np.sum(w)
                )
                e_angles["pos_y_spread"] = np.sqrt(
                    np.sum(w * (pos_y - e_angles["pos_y"]) ** 2) / np.sum(w)
                )
                e_angles["pos_z_spread"] = np.sqrt(
                    np.sum(w * (pos_z - e_angles["pos_z"]) ** 2) / np.sum(w)
                )


@numba.njit(cache=True, nogil=True)
def get_average_angle(
    hitlets_in_event: numba.typed.typedlist.List,
    pmt_properties: np.ndarray,
    start_channel: int = 2000,
) -> np.ndarray:
    """Computes azimuthal angle as an area weighted mean over all hitlets.

    :param hitlets_in_event: numba.typed.List containing the hitlets per event.
    :param pmt_properties: numpy.sturctured.array containing the PMT positions in the fields "x" and
        "y".
    :param start_channel: First channel e.g. 2000 for nevto.
    :return: np.array holding the azimuthal angles.

    """
    res = np.zeros(len(hitlets_in_event), np.float32)
    for ind, hitlets in enumerate(hitlets_in_event):
        if np.sum(hitlets["area"]):
            x = pmt_properties["x"][hitlets["channel"] - start_channel]
            y = pmt_properties["y"][hitlets["channel"] - start_channel]

            weighted_mean_x = np.sum(x * hitlets["area"]) / np.sum(hitlets["area"])
            weighted_mean_y = np.sum(y * hitlets["area"]) / np.sum(hitlets["area"])
            res[ind] = _circ_angle(weighted_mean_x, weighted_mean_y)
        else:
            res[ind] = np.nan
    return res


@numba.njit(cache=True, nogil=True)
def circ_angle(x_values: np.ndarray, y_values: np.ndarray) -> np.ndarray:
    """Loops over a set of x and y values and computes azimuthal angle.

    :param x_values: x-coordinates
    :param y_values: y-coordinates
    :return: angles

    """
    res = np.zeros(len(x_values), dtype=np.float32)
    for ind, (x, y) in enumerate(zip(x_values, y_values)):
        res[ind] = _circ_angle(x, y)
    return res


@numba.njit(cache=True, nogil=True)
def _circ_angle(x: float, y: float) -> float:
    if x > 0 and y >= 0:
        # 1st quadrant
        angle = np.abs(np.arctan(y / x))
        return angle
    elif x <= 0 and y > 0:
        # 2nd quadrant
        angle = np.abs(np.arctan(x / y))
        return np.pi / 2 + angle
    elif x < 0 and y <= 0:
        # 3rd quadrant
        angle = np.abs(np.arctan(y / x))
        return np.pi + angle
    elif y < 0 and x >= 0:
        # 4th quadrant
        angle = np.abs(np.arctan(x / y))
        return 3 / 2 * np.pi + angle
    elif x == 0 and y == 0:
        return np.nan
    else:
        print(x, y)
        raise ValueError("It should be impossible to arrive here, but somehow we managed.")


@numba.njit(cache=True, nogil=True)
def first_hitlets(
    hitlets_per_event: np.ndarray, max_time: int
) -> ty.Tuple[numba.typed.List, np.ndarray]:
    """Returns hitlets within the first "max_time" ns of an event.

    :param hitlets_per_event: numba.typed.List of hitlets per event.
    :param max_time: int max allowed time difference to leading hitlet in ns.

    """
    res_hitlets_in_event = numba.typed.List()
    res_n_prompt = np.zeros(len(hitlets_per_event), np.int16)
    for ind, hitlets in enumerate(hitlets_per_event):
        m = (hitlets["time"] - hitlets[0]["time"]) < max_time
        h = hitlets[m]
        res_hitlets_in_event.append(h)
        res_n_prompt[ind] = len(h)
    return res_hitlets_in_event, res_n_prompt



"""Some shared defaults."""

DEFAULT_POSREC_ALGO = "cnf"

HE_PREAMBLE = """High energy channels: attenuated signals of the top PMT-array\n"""

MV_PREAMBLE = "Muno-Veto Plugin: Same as the corresponding nVETO-PLugin.\n"

NV_HIT_DEFAULTS = {
    "save_outside_hits_nv": (3, 15),
    "hit_min_amplitude_nv": (
        "list-to-array://"
        "pad-array://"
        "xedocs://hit_thresholds"
        "?pad_left=2000"
        "&as_list=True"
        "&sort=pmt"
        "&attr=value"
        "&detector=neutron_veto"
        "&run_id=plugin.run_id"
        "&version=ONLINE"
    ),
}

MV_HIT_DEFAULTS = {
    "save_outside_hits_mv": (2, 5),
    "hit_min_amplitude_mv": (
        "list-to-array://"
        "pad-array://"
        "xedocs://hit_thresholds"
        "?pad_left=1000"
        "&as_list=True"
        "&sort=pmt"
        "&attr=value"
        "&detector=muon_veto"
        "&run_id=plugin.run_id"
        "&version=ONLINE"
    ),
}

FAR_XYPOS_S2_TYPE = 20
WIDE_XYPOS_S2_TYPE = 22



import argparse
from ast import literal_eval
import json
import os
import os.path as osp

import numpy as np
from tqdm import tqdm

import strax
import straxen


def main():
    parser = argparse.ArgumentParser(description="Refresh strax raw_records created with < v0.9.0")
    parser.add_argument("--parent_folder", default=".", help="Folder with raw_records folders.")
    parser.add_argument(
        "--no_run_metadata",
        action="store_true",
        help=(
            "Refresh even if you have lost the run metadata. "
            "Some useful sanity checks will be disabled."
        ),
    )
    parser.add_argument(
        "--procrustean",
        action="store_true",
        help=(
            "Delete records that fall across chunk boundaries. "
            "Use as a last resort for triggerless data you do not want "
            "to reconvert from pax properly."
        ),
    )
    parser.add_argument(
        "--xnt",
        action="store_true",
        help="Set if records came from the DAQReader, not RecordsFromPax",
    )
    parser.add_argument("run_id", help="run_id to convert")
    args = parser.parse_args()
    run_id = args.run_id
    parent_folder = args.parent_folder

    ##
    # Prepare context
    ##
    st = strax.Context(
        storage=[
            strax.DataDirectory(
                parent_folder,
                # We WILL overwrite your data
                # just not through the usual means:
                readonly=True,
            )
        ],
        **straxen.contexts.common_opts,
    )
    if args.xnt:
        st.register(straxen.DAQReader)
    else:
        st.register(straxen.RecordsFromPax)

    ##
    # Get metadata
    ##
    folder = st.storage[0].find(st.key_for(run_id, "raw_records"), fuzzy_for="raw_records")[1]
    md = st.get_metadata(run_id, "raw_records")
    metadata_fn = os.path.join(folder, strax.RUN_METADATA_PATTERN % strax.dirname_to_prefix(folder))
    assert osp.exists(metadata_fn)
    dtype = np.dtype(literal_eval(md["dtype"]))
    record_length = strax.record_length_from_dtype(dtype)

    if not args.no_run_metadata:
        run_md = st.run_metadata(run_id)
        run_start, run_end = [
            int(x.timestamp()) * int(1e9) for x in [run_md["start"], run_md["end"]]
        ]
    else:
        run_start, run_end = None, None

    if not len(md["chunks"]):
        raise ValueError("Cannot convert data: no chunks!")
    if "start" in md["chunks"][0]:
        raise ValueError("This data was already converted")

    ##
    # Convert data
    ##
    last_endtime = 0
    for i, c in enumerate(tqdm(md["chunks"], desc=f"Converting raw_records for {run_id}")):
        filename = osp.join(folder, c["filename"])
        rr = strax.load_file(filename, dtype=dtype, compressor=md["compressor"])

        if not len(rr):
            raise ValueError("Cannot convert data with empty chunks")

        if rr[0]["time"] < last_endtime:
            if args.procrustean:
                to_cut = rr["time"] < last_endtime
                print(f"[!!] Removing {to_cut.sum()} records from chunk {i} to remove overlaps!")
                rr = rr[~to_cut]
            else:
                raise ValueError(
                    f"Cannot convert data: chunk {i}'s data starts "
                    f"at {rr[0]['time']} while the previous chunk's data "
                    f"ended at {last_endtime}"
                )

        if i == 0:
            if run_start is not None:
                c["start"] = run_start
            else:
                c["start"] = rr[0]["time"]
        else:
            c["start"] = last_endtime
        c["end"] = last_endtime = strax.endtime(rr).max()

        new_rr = np.zeros(len(rr), dtype=strax.raw_record_dtype(record_length))
        strax.copy_raw_records(rr, new_rr)
        if "baseline" in rr.dtype.fields:
            # Undo baselining
            new_rr["data"] = rr["baseline"][:, np.newaxis] - rr["data"]
            strax.zero_out_of_bounds(new_rr)

        c["run_id"] = run_id
        c["nbytes"] = new_rr.nbytes
        c["filesize"] = strax.save_file(filename, new_rr, compressor=md["compressor"])
        # We must rewrite these too, the chunk count could have changed
        # if args.procrustean
        c["n"] = len(rr)
        c["first_time"] = rr[0]["time"]
        c["first_endtime"] = strax.endtime(rr[0])
        c["last_time"] = rr[-1]["time"]
        c["last_endtime"] = strax.endtime(rr[-1])

    if run_start is None:
        run_start = md["chunks"][0]["start"]
        run_end = md["chunks"][-1]["start"]

    ##
    # Set and write out new metadata
    ##
    md["start"] = run_start
    md["end"] = run_end
    md["run_id"] = run_id
    md["data_kind"] = "raw_records"
    md["converted_from_old_strax"] = md["strax_version"]
    md["strax_version"] = strax.__version__
    md["dtype"] = np.dtype(strax.raw_record_dtype(record_length)).descr.__repr__()

    with open(metadata_fn, mode="w") as f:
        f.write(json.dumps(md, sort_keys=True, indent=4, cls=strax.NumpyJSONEncoder))


if __name__ == "__main__":
    main()



import numpy as np
import strax
import straxen

export, __all__ = strax.exporter()


@export
class IndividualPeakMonitor(strax.Plugin):
    """Plugin to write data needed for the online SE monitor to the online- monitor collection in
    the runs-database. Data that is written by this plugin should be small such as to not overload
    the runs- database. If the peaks are large, random max_bytes of data are selected from the
    peaks.

    This plugin takes 'peak_basics' and 'peak_positions_mlp'. Although they are not strictly
    related, they are aggregated into a single data_type in order to minimize the number of
    documents in the online monitor.

    Produces 'individual_peak_monitor' with info on the peaks and their positions.

    """

    online_max_bytes = straxen.URLConfig(
        default=6e6, track=True, help="Maximum amount of bytes of data for MongoDB document"
    )

    depends_on = ("peak_basics", "peak_positions_mlp")
    provides = "individual_peak_monitor"
    data_kind = "individual_peak_monitor"
    __version__ = "0.0.1"

    def infer_dtype(self):
        dtype = [
            (("Peak integral in PE", "area"), np.float32),
            (("Reconstructed mlp peak x-position", "x_mlp"), np.float32),
            (("Reconstructed mlp peak y-position", "y_mlp"), np.float32),
            (("Width (in ns) of the central 50% area of the peak", "range_50p_area"), np.float32),
            (("Fraction of original peaks array length that is saved", "weight"), np.float32),
        ] + strax.time_fields
        return dtype

    def compute(self, peaks):
        peaks_size = peaks.nbytes

        if peaks_size > self.online_max_bytes:
            # Calculate fraction of the data that can be kept
            # to reduce datasize
            new_len = int(len(peaks) / peaks_size * self.online_max_bytes)
            idx = np.random.choice(np.arange(len(peaks)), replace=False, size=new_len)
            data = peaks[strax.stable_sort(idx)]

        else:  # peaks_size <= self.max_bytes:
            data = peaks
        res = np.zeros(len(data), dtype=self.dtype)
        res["time"] = data["time"]
        res["x_mlp"] = data["x_mlp"]
        res["y_mlp"] = data["y_mlp"]
        res["area"] = data["area"]
        res["range_50p_area"] = data["range_50p_area"]
        res["endtime"] = data["endtime"]

        if len(data):
            res["weight"] = len(peaks) / len(data)
        else:
            res["weight"] = 0

        return res



"""
AJAX: XENON-nT
Aggregate Junking Ancient Xenon-data
cleaning tool to remove old data from event builders.
=============================================
Joran Angevaare, 2020

This tool keeps the event builders clean by performing any of the cleaning modes
(try 'ajax --help' for a complete description).

Some of these modes also affect the live_data rather than the processed data. As there can
be multiple instances of ajax running there is a dedicated machine 'eb_can_clean_ceph'
that may delete the live_data such that we prevent multiple attempts to perform a single
action.
"""

__version__ = "0.5.0"

import argparse
import warnings
from datetime import datetime, timedelta
import logging
import os
import socket
import shutil
import pytz
import threading
import time
import re
import numpy as np
import straxen
from straxen import daq_core
from straxen.daq_core import ceph_folder, output_folder, pre_folder, non_registered_folder
import strax


ajax_thresholds = {
    # Remove the live data only if this many seconds old
    "remove_live_after": 24 * 3600,  # s
    # Remove do not remove successfully processed runs if they have finished less than
    # this many seconds ago to e.g. allow for saving et cetera
    "wait_after_processing": 12 * 3600,  # s
    # Remove high level plugins if the run is this old
    # TODO: change the value to a different one
    "remove_high_level_data": 365 * 24 * 3600,  # s
    # Minimum time for restarting ajax to check for new stuff to clean (essentially the
    # timeout)
    "nap_time": 3600,  # s
    # Short nap time
    "short_nap": 5,  # s
}

# The low level data is taken care of by admix so we don't have to delete this. All other
# data-types are deleted in the clean_high_level_data routine
low_level_data = [
    "raw_records*",
    "records*",
    "veto_regions",
    "pulse_counts",
    "led_calibration",
    "peaklets",
]

# Open deletion threads with this prefix (and also look for these before quiting ajax)
thread_prefix = "ajax_delete"

# To prevent multiple ebs from cleaning ceph only this one can actually do it
eb_can_clean_ceph = "eb0.xenon.local"


def main():
    print(f"---\n ajax version {__version__}\n---")
    parser = argparse.ArgumentParser(description="XENONnT cleaning manager")
    parser.add_argument(
        "--force", action="store_true", help="Forcefully remove stuff from this host"
    )
    parser.add_argument(
        "--ask_confirm",
        action="store_true",
        help="Always ask for confirmation before deleting data/updating the rundoc",
    )
    parser.add_argument("--delete_live", action="store_true", help="delete live data for this run")
    parser.add_argument(
        "--production",
        action="store_true",
        help="Execute the deletion commands. If not specified, ajax assumes you want to test",
    )
    parser.add_argument("--logging", default="DEBUG", help="logging level (DEBUG/INFO/WARNING)")

    actions = parser.add_mutually_exclusive_group()
    actions.add_argument(
        "--number",
        type=int,
        metavar="NUMBER",
        help="Process a single run, regardless of its status.",
    )
    actions.add_argument(
        "--clean",
        type=str,
        help="Run ajax in any of the following modes: clean [ceph, "  # noqa
        "unregistered, abandoned, high_level_data, all]\n"
        '"ceph": Remove successfully processed runs and abandoned runs from /live_data\n'
        '"unregistered": remove all data from this host that is not registered in the rundb\n'
        '"abandoned": remove all the data associated to abandoned runs\n'
        '"high_level_data": remove all high level data on this host\n'
        '"non_latest": remove data on this host if it was not the last to process a given run \n'
        '"database": check if all the entries that the database claims are actually here \n'
        '"old_hash": remove high level data if the hash doesnt equal the latest for this'
        " datatype \n"
        '"production": clean the production folder (bad idea if restrax is running since it writes'
        " there) \n"
        '"all": Clean everything that AJAX can get its hands on: unregistered data, high level '
        "data",
    )

    args = parser.parse_args()
    hostname = socket.getfqdn()
    databases = daq_core.DataBases(production=args.production)

    if not hasattr(logging, args.logging):
        raise AttributeError(f"Set --logging to a logging level like DEBUG or INFO")
    logging_level = getattr(logging, args.logging)
    try:
        import daqnt

        log_name = "ajax_" + hostname + ("" if args.production else "_TESTING")
        log = daqnt.get_daq_logger("main", log_name, level=logging_level)
    except ModuleNotFoundError:
        logging.basicConfig(
            level=logging_level,
            format="%(asctime)s %(name)s %(levelname)-8s %(message)s",
            datefmt="%m-%d %H:%M",
        )
        log = logging.getLogger()

    if args.delete_live:
        if not args.number:
            raise ValueError("Specify which number with --number")
    if args.force:
        log.warning(
            f"main::\tDANGER ZONE you are forcefully deleting data that may "
            f"result in an irrecoverable loss of data."
        )
        log.info(
            f"main::\tPlease note that execute argument is {args.production} which "
            f"means you are {'' if not args.production else '!NOT!'} safe"
        )
        if not args.ask_confirm:
            raise NotImplementedError(
                f"main::\tI cannot let your forcefully delete data without asking for "
                f"confirmation. Add --ask_confirm. Bye, bye"
            )

        if not input("Want to proceed willingly? [y]").lower() == "y":
            log.info(f"main::\tAlright no unsafe operations, bye bye")
            exit(-1)
        if args.clean != "abandoned" and not args.number:
            raise NotImplementedError("main::\tI don't want to have this option enabled (yet).")

    if args.clean in ["all", "old_hash"]:
        # Need the context for the latest hash
        st = straxen.contexts.xenonnt_online()

    def log_warning(message, priority="warning", run_id=None):
        getattr(log, priority.lower())(message)
        databases.log_warning(
            message,
            priority=priority,
            run_id=run_id,
            production=args.production,
            user=f"ajax_{hostname}",
        )

    for f in (output_folder, non_registered_folder):
        if os.access(f, os.W_OK) is not True:
            log_warning(f"main::\tNo writing access to {f}", priority="fatal")
            raise IOError(f"main::\tNo writing access to {f}")

    # Runs database
    run_db = databases.run_db
    run_coll = databases.run_coll
    run_db.command("ping")

    # DAQ database
    daq_db = databases.daq_db
    bs_coll = databases.bs_coll
    daq_db.command("ping")

    def wait_on_delete_thread():
        """Check that the threads with the thread_prefix are finished before continuing."""
        threads = threading.enumerate()
        for thread in threads:
            if thread_prefix in thread.name:
                wait = True
                while wait:
                    wait = False
                    if thread.isAlive():
                        log.info(
                            f"{thread.name} still running take a "
                            f"{ajax_thresholds['short_nap']} s nap"
                        )
                        time.sleep(ajax_thresholds["short_nap"])
                        wait = True
        log.info(f"wait_on_delete_thread::\tChecked that all {thread_prefix}* finished")

    def set_state(state):
        """Inform the bootstrax collection we're in a different state.

        if state is None, leave state unchanged, just update heartbeat time

        """
        ajax_state = dict(
            host="ajax." + hostname,
            pid=os.getpid(),
            time=now(),
            state=state,
            mode=f"clean {args.clean}",
            production_mode=args.production,
        )
        bs_coll.insert_one(ajax_state)

    def run():
        """Main function."""
        if args.number:
            remove_run_from_host(
                args.number, delete_live=args.delete_live, force=args.force, reason="manual"
            )
            wait_on_delete_thread()
            return
        if not args.clean:
            raise ValueError("either --number or --clean should be specified")

        _modes = {
            "ceph": clean_ceph,
            "high_level_data": clean_high_level_data,
            "non_latest": clean_non_latest,
            "unregistered": clean_unregistered,
            "production": clean_production_folder,
            "abandoned": clean_abandoned,
            "old_hash": clean_old_hash,
            "database": clean_database,
        }

        if args.clean not in _modes and args.clean != "all":
            raise ValueError(f"Unknown cleaning mode {args.clean}")

        if args.clean != "all":
            # Do this mode
            _modes[args.clean]()
            if hostname == eb_can_clean_ceph and (args.clean in ["abandoned", "database"]):
                _modes[args.clean](delete_live=True)
            log.info(f"Done with {args.clean}")
            return

        # Apparently we want to clean everything. Let's loop:
        while True:
            try:
                if hostname == eb_can_clean_ceph:
                    clean_ceph()
                    clean_abandoned(delete_live=True)

                    # Don't do clean_unregistered by default (only if specified
                    # as specific argument on host "eb_can_clean_ceph") as
                    # someone might be storing some test data in the /live_data
                    # folder. That is bad practise though
                    # clean_unregistered(delete_live=True)

                clean_abandoned()
                clean_non_latest()

                # These two modes below shouldn't be done on autopilot at the
                # time of writing.
                # clean_unregistered()
                # clean_database()
                if not args.production:
                    break
                log.info(f'Loop finished, take a {ajax_thresholds["nap_time"]} s nap')
                time.sleep(ajax_thresholds["nap_time"])
            except (KeyboardInterrupt, SystemExit, NotImplementedError) as e:
                log.info("\nStopping, wait a second for the delete threads\n")
                wait_on_delete_thread()
                raise e
            except Exception as fatal_error:
                log.error(
                    f"Fatal warning:\tran into {fatal_error}. Try logging error and restart ajax"
                )
                try:
                    log_warning(f"Fatal warning:\tran into {fatal_error}", priority="error")
                except Exception as warning_error:
                    log.error(f"Fatal warning:\tcould not log {warning_error}")
                # This usually only takes a minute or two
                time.sleep(60)
                log.warning("Restarting main loop")

    try:
        set_state("idle")
        run()
    except (KeyboardInterrupt, SystemExit) as e:
        log.info("\nStopping, wait a second for the delete threads\n")
        wait_on_delete_thread()
        raise e

    wait_on_delete_thread()
    log.info(f"main::\tAjax finished, bye bye")

    def clean_ceph():
        """Look for old data on ceph.

        If found, delete it. Recursively iterate until no old data is found. This function only
        works on the the host that is eb_can_clean_ceph.

        """
        set_state("clean_ceph")
        if not hostname == eb_can_clean_ceph:
            log.info(f"for cleaning ceph, go to {eb_can_clean_ceph}")
            return

        rd = run_coll.find_one(
            {
                "bootstrax.state": "done",
                "status": "transferred",
                "start": {"$lt": now(-ajax_thresholds["remove_live_after"])},
                "bootstrax.time": {"$lt": now(-ajax_thresholds["wait_after_processing"])},
                "data.type": "live",
            }
        )
        if rd is None:
            return
        else:
            run_number = rd["number"]
            log.info(f"remove data associated to run {run_number}")
            remove_run_from_host(
                run_number, delete_live=True, force=args.force, reason="clean ceph"
            )
            log.info(f"finished for {run_number}, take a {ajax_thresholds['short_nap']} s nap")
            time.sleep(ajax_thresholds["short_nap"])

            # Repeat.
            if args.production:
                wait_on_delete_thread()
                clean_ceph()

        # Finally, check that there is no old data on ceph that is also not in the rundoc.
        # clean_unregistered(delete_live=True)

    def clean_high_level_data():
        """Check the runs database for old data on this host and clean all non-low-level data-types
        from this host."""
        set_state("clean_high_level_data")
        raise NotImplementedError(
            "This is a drastic measure that we do not intend using at the time of writing"
        )

        # We need to grep all the rundocs at once as simply doing one at the time (find_one)
        # might get stuck as the query may yield the same result the next time it is called.
        rds = run_coll.find(
            {
                "bootstrax.state": "done",
                "status": "transferred",
                "start": {"$lt": now(-ajax_thresholds["remove_high_level_data"])},
                "data.host": hostname,
            }
        )
        if not rds:
            # Great, we are clean
            return

        for rd in rds:
            # First check that we actually have raw_records stored somewhere
            _, have_raw_records = check_rundoc_for_live_and_raw(rd)
            if not have_raw_records:
                break

            # Okay, good to go let's see if it is high level data and if so, delete it
            for ddoc in rd["data"]:
                # Only delete data on this host
                if "host" in ddoc and ddoc["host"] == hostname:
                    is_low_level = re.findall("|".join(low_level_data), ddoc["type"])
                    if is_low_level:
                        continue
                    loc = ddoc["location"]
                    if "raw_records" in ddoc["type"]:
                        raise ValueError("your regex syntax fails!")
                    elif os.path.exists(loc):
                        log.info(f"delete data at {loc}")
                        delete_data(
                            rd,
                            loc,
                            ddoc["type"],
                            test=not args.production,
                            reason="high level data",
                        )
                    else:
                        loc = loc + "_temp"
                        log.info(f"delete data at {loc}")
                        delete_data(
                            rd,
                            loc,
                            ddoc["type"],
                            test=not args.production,
                            reason="high level data",
                        )

    def clean_non_latest():
        """Remove data on this host if the boostrax.host field is not this host while processing has
        finished."""
        set_state("clean_non_latest")
        # We need to grep all the rundocs at once as simply doing one at the time (find_one)
        # might get stuck as the query may yield the same result the next time it is called.
        rds = run_coll.find(
            {
                "bootstrax.state": "done",
                "status": "transferred",
                "bootstrax.time": {"$lt": now(-ajax_thresholds["wait_after_processing"])},
                "bootstrax.host": {"$ne": hostname},
                "data.host": hostname,
            }
        )
        if not rds:
            # Great, we are clean
            return

        for rd in rds:
            # First check that we actually have raw_records stored on one of the other ebs
            have_raw_records = False
            for dd in rd["data"]:
                if dd["type"] == "raw_records" and dd["host"] != hostname:
                    have_raw_records = True
                    break
            if not have_raw_records:
                break

            # Okay, good to go let's see if it is high level data and if so, delete it
            for ddoc in rd["data"]:
                # Only delete data on this host
                if "host" in ddoc and ddoc["host"] == hostname:
                    loc = ddoc["location"]
                    if os.path.exists(loc):
                        log.info(f"clean_non_latest::\tdelete data at {loc}")
                        delete_data(
                            rd, loc, ddoc["type"], test=not args.production, reason="non latest"
                        )
                    else:
                        loc = loc + "_temp"
                        log.info(f"clean_non_latest::\tdelete data at {loc}")
                        delete_data(
                            rd, loc, ddoc["type"], test=not args.production, reason="non latest"
                        )

    def clean_production_folder():
        warnings.warn(
            "Running clean of production folder is a bad idea if you "
            "are running restrax at the same time (restrax does not "
            "register data while writing it)"
        )
        for _ in strax.utils.tqdm(list(range(30)), desc="Continue in 30 seconds"):
            pass
        clean_unregistered(delete_live=False, _clean_from=output_folder)

    def clean_unregistered(delete_live=False, _clean_from=pre_folder):
        """Clean data that is not in the database. To do this check the output folder and remove
        files if there is no corresponding entry in the rundoc.

        :param delete_live: bool, if true delete unregistered live data.
        :param _clean_from: which production (not live) folder to clean from

        """
        set_state("clean_unregistered")
        # We should only do this for the pre-folder since the production folder does
        # have unregistered data when we are still trying to recompress the data
        folder_to_check = _clean_from if delete_live is False else ceph_folder
        all_data = os.listdir(folder_to_check)
        run_ids = []
        for f in all_data:
            run_ids.append(f.split("-")[0])
        run_ids = np.unique(run_ids)
        log.info(
            f"clean_unregistered::\tfound {len(run_ids)} runs stored on"
            f"{folder_to_check}. Checking that each is in the runs-database"
        )
        for run_id in run_ids:
            run_number = int(run_id)
            remove_if_unregistered(run_number, delete_live=delete_live, clean_from=_clean_from)

    def clean_old_hash():
        """Loop over the files on the host and check that the lineage is the same as the current
        lineage hash we would get from straxen."""
        set_state("clean_old_hash")
        files = os.listdir(output_folder)
        for f in files:
            run_id, data_type, lineage_hash = f.split("-")
            if bool(re.findall("|".join(low_level_data), data_type)) or "_temp" in lineage_hash:
                continue
            elif data_type not in st._plugin_class_registry:
                log.warning(f"{data_type} is not registered!")
                continue

            current_st_hash = st.key_for(run_id, data_type).lineage_hash
            if current_st_hash != lineage_hash:
                loc = os.path.join(output_folder, f)
                log.info(
                    f"clean_old_hash::\tLineage for run {run_id}, {data_type} is "
                    f"{current_st_hash}. Removing old hash: {lineage_hash} from {loc}."
                )
                rd = run_coll.find_one(
                    {
                        "bootstrax.state": "done",
                        "status": "transferred",
                        "bootstrax.time": {"$lt": now(-ajax_thresholds["wait_after_processing"])},
                        "data.type": data_type,
                        "data.host": hostname,
                        "number": int(run_id),
                    }
                )
                if "raw_records" in data_type:
                    raise ValueError(
                        f"You did some sloppy regex on the data-type, almost deleted {data_type}!"
                    )
                if rd:
                    delete_data(
                        rd, loc, data_type, test=not args.production, reason="old lineage hash"
                    )
                else:
                    log.info(
                        f"{data_type} at {loc} not registered as done for {run_id}. "
                        "Perhaps the data still processing the run?"
                    )

    def clean_abandoned(delete_live=False):
        """Recursively delete data associated to abandoned runs. If deleting live data, submit
        multiple threads at the same time.

        :param delete_live: bool, if true also delete the live_data of these runs.

        """
        set_state("clean_abandoned")
        # Notice that we thread the deletion of live_data. As such we need to allow multiple
        # runs to be deleted simultaneously.
        if not delete_live:
            rd = run_coll.find_one({"bootstrax.state": "abandoned", "data.host": hostname})
            if rd is None:
                log.info("clean_abandoned::\tNo more matches in rundoc")
                return
            else:
                # Make it iterable for the loop below
                rds = [rd]
        else:
            rds = run_coll.find(
                {
                    "bootstrax.state": "abandoned",
                    "bootstrax.time": {"$lt": now(-ajax_thresholds["wait_after_processing"])},
                    "data.location": "/live_data/xenonnt",
                    "data.host": "daq",
                }
            )

        # Count the number of threads. Only allow one if we aks for user input.
        i = 0
        i_max = 1 if args.ask_confirm else 5
        for rd in rds:
            if i >= i_max:
                break
            run_number = rd["number"]
            log.info(f"clean_abandoned::\tremove data associated to run {run_number}")
            # Please note that we have to force these runs always since they should not be
            # stored elsewhere. They are abandoned for a reason!
            remove_run_from_host(
                run_number, delete_live=delete_live, force=True, reason="run abandonned"
            )
            i += 1
        if not i:
            log.info("clean_abandoned::\tNo more live_data matches in rundoc")
            # Apparently there is no rd in rds
            return

        # Repeat
        if args.production:
            if delete_live:
                wait_on_delete_thread()
            return clean_abandoned(delete_live=delete_live)

    def clean_database(delete_live=False):
        """Remove entries from the database if the data is not on this host."""
        set_state("clean_database")
        # We need to grep all the rundocs at once as simply doing one at the
        # time (find_one) might get stuck as the query may yield the same
        # result the next time it is called.
        if not delete_live:
            rds = run_coll.find(
                {
                    "bootstrax.state": "done",
                    "bootstrax.time": {"$lt": now(-ajax_thresholds["wait_after_processing"])},
                    "data.host": hostname,
                }
            )
        else:
            rds = run_coll.find(
                {
                    "bootstrax.state": "done",
                    "bootstrax.time": {"$lt": now(-ajax_thresholds["wait_after_processing"])},
                    "data.host": "daq",
                }
            )
        if not rds:
            # Great, we are clean
            return

        for rd in rds:
            # Okay, good to go let's see if it is high level data and if so, delete it
            for ddoc in rd["data"]:
                # Only delete data on this host
                if ("host" in ddoc and ddoc["host"] == hostname) or (
                    "host" in ddoc and ddoc["host"] == "daq" and delete_live
                ):
                    loc = ddoc["location"]
                    if not os.path.exists(loc):
                        log.info(
                            "clean_database::\tdelete entry of data from "
                            f"{rd['number']} at {loc} as it does not exist"
                        )
                        delete_data(
                            rd,
                            loc,
                            ddoc["type"],
                            test=not args.production,
                            reason="reason already removed from eb?!",
                        )

    def _rmtree(path):
        """Wrapper for shutil.rmtree. All deletion statements in this script go through this
        function in order to make sure that the args.production statement is always double (tripple)
        checked before deleting data.

        :param path: path to delete

        """
        if args.production:
            if confirm(f"delete {path}?"):
                shutil.rmtree(path)
        else:
            log.info(f"TESTING:\tshutil.rmtree({path})")
            if not os.path.exists(path):
                raise ValueError(f"{path} does not exist")

    def threaded_delete_data(
        rd, path, data_type, test=True, ignore_low_data_check=False, reason=""
    ):
        """Wrapper for delete_data to run in separate threads.

        :param rd: rundoc
        :param path: location of the folder to be deleted
        :param data_type: type of data to be deleted
        :param test: bool if we are testing or not. If true, nothing will be deleted.
        :param ignore_low_data_check: ignore the fact that this might be the only copy of the data.
            We can specify this e.g. if we know this is data associated to some abandoned run.

        """

        thread_name = thread_prefix + path.split("/")[-1]
        delete_thread = threading.Thread(
            name=thread_name,
            target=delete_data,
            args=(rd, path, data_type, test, ignore_low_data_check),
            kwargs={"reason": reason},
        )
        log.info(f"Starting thread to delete {path} at {now()}")
        # We rather not stop deleting the live_data if something else fails. Set the thread
        # to daemon.
        delete_thread.setDaemon(True)
        delete_thread.start()
        log.info(
            f"DeleteThread {path} should be running in parallel, continue MainThread now: {now()}"
        )

    def delete_data(
        rd,
        path,
        data_type,
        test=True,
        ignore_low_data_check=False,
        reason="",
    ):
        """Delete data and update the rundoc.

        :param rd: rundoc
        :param path: location of the folder to be deleted
        :param data_type: type of data to be deleted
        :param test: bool if we are testing or not. If true, nothing will be deleted.
        :param ignore_low_data_check: ignore the fact that this might be the only copy of the data.
            We can specify this e.g. if we know this is data associated to some abandoned run.

        """
        # First check that we are not deleting essential data (unless we are working
        # with abandoned runs or so.
        if not ignore_low_data_check:
            n_live, n_rr = check_rundoc_for_live_and_raw(rd)
            if not (n_live or n_rr >= 1 + int("raw_records" in data_type)):
                message = (
                    f"Trying to delete {data_type} but we only have {n_live}"
                    f" live- and {n_rr} raw_record-files in the "
                    "runs-database. This might be an essential copy of the"
                    " data!"
                )
                log_warning(message, priority="fatal", run_id=f'{rd["number"]:06}')
                if not test:
                    raise ValueError(message)

        if os.path.exists(path):
            log.info(f"Deleting data at {path}")
            if not test:
                _rmtree(path)
            log.info(f"deleting {path} finished")
        else:
            log.info(f"There is no data on {path}! Just doing the rundoc.")

        # Remove the data location from the rundoc and append it to the 'deleted_data' entries
        if not os.path.exists(path):
            log.info("changing data field in rundoc")
            for ddoc in rd["data"]:
                if ddoc["type"] == data_type and ddoc["host"] in ("daq", hostname):
                    break
            for k in ddoc.copy().keys():
                if k in ["location", "meta", "protocol"]:
                    ddoc.pop(k)

            ddoc.update({"at": now(), "by": f"ajax.{hostname}", "reason": reason})
            log.info(f"update with {ddoc}")
            if args.production and not test:
                if confirm("update rundoc?"):
                    run_coll.update_one(
                        {"_id": rd["_id"]},
                        {
                            "$addToSet": {"deleted_data": ddoc},
                            "$pull": {
                                "data": {"type": data_type, "host": {"$in": ["daq", hostname]}}
                            },
                        },
                    )
            else:
                log.info(f"Update ddoc with : {ddoc}")
        elif not test and not args.ask_confirm:
            raise ValueError(f"Something went wrong we wanted to delete {path}!")

    def check_rundoc_for_live_and_raw(rd):
        """Count the number of files of live_data (cannot be >1) and raw_data (transfers can get it
        greater than 1)

        :param rd: rundoc
        :return: length 2 tuple with n_live_data and n_raw_records being the number of files in the
            rundoc for the live data and raw records respectively.

        """
        n_live_data, n_raw_records = 0, 0
        for dd in rd["data"]:
            if dd["type"] == "live":
                n_live_data += 1
            if dd["type"] == "raw_records":
                n_raw_records += 1
        return n_live_data, n_raw_records

    def remove_run_from_host(number, delete_live=False, force=False, reason=""):
        """Save way of removing data from host if data registered elsewhere.

        :param number: run number (not ID!)
        :param delete_live: bool, if true delete the live_data else the processed data
        :param force: forcefully remove the data even if we don't have the right copies (e.g.
            deleting /live_data when the raw_records are not stored. Be careful with this option!
            Should only be used for the deletion of abandoned runs.

        """
        # Query the database to remove data
        rd = run_coll.find_one(
            {"number": number, "data.host": hostname if not delete_live else "daq"}
        )
        if not rd:
            log_warning(
                f"No registered data for {number} on {hostname}",
                run_id=f"{number:06}",
                priority="info",
            )
            return

        have_live_data, have_raw_records = check_rundoc_for_live_and_raw(rd)

        for ddoc in rd["data"]:
            # This is processed data on the eventbuilders
            if "host" in ddoc and ddoc["host"] == hostname:
                if delete_live:
                    # If you want to delete the live data you shouldn't consider this ddoc
                    continue
                loc = ddoc["location"]
                if not force and not have_live_data and "raw_records" in ddoc["type"]:
                    # If we do not have live_data, don't delete raw_records. However, if we
                    # --force deletion, do go to the next else statement
                    log.info(
                        f"prevent {loc} from being deleted. The live_data has already been removed"
                    )
                else:
                    log.info(f"delete data at {loc}")
                    delete_data(
                        rd,
                        loc,
                        ddoc["type"],
                        test=not args.production,
                        ignore_low_data_check=force,
                        reason=reason,
                    )

                    loc = loc + "_temp"
                    if os.path.exists(loc):
                        log.info(f"delete data at {loc}")
                        delete_data(
                            rd,
                            loc,
                            ddoc["type"],
                            test=not args.production,
                            ignore_low_data_check=force,
                            reason=reason,
                        )
            elif "host" in ddoc and ddoc["host"] == "daq":
                # This is the live_data
                if not delete_live:
                    log.info(
                        f'prevent {ddoc["location"]} from being deleted. Do so with --delete_live'
                    )
                    # If you want to delete processed data you shouldn't consider this ddoc
                    continue

                run_id = "%06d" % number
                loc = os.path.join(ddoc["location"], run_id)
                if not force and not have_raw_records:
                    # If we do not have raw records, don't delete this data. However, if we
                    # --force deletion, do go to the next else statement
                    log_warning(
                        f"Unsafe to delete {loc}, no raw_records registered. Force with --force",
                        priority="info",
                    )
                elif not force and have_raw_records:
                    log.info(f"Deleting {loc} since we have raw_records registered.")
                    threaded_delete_data(
                        rd,
                        loc,
                        ddoc["type"],
                        test=not args.production,
                        ignore_low_data_check=False,
                        reason=reason,
                    )
                # Redundant elif but let's double check the force nonetheless.
                elif force:
                    log.info(f"Forcefully delete {loc}, but no raw_records registered!")
                    threaded_delete_data(
                        rd,
                        loc,
                        ddoc["type"],
                        test=not args.production,
                        ignore_low_data_check=True,
                        reason=reason,
                    )

    def remove_if_unregistered(number, delete_live=False, clean_from=pre_folder):
        """Given a run number delete data on this machine that matches that number.

        :param number: int! the run_number (not run_id)
        :param delete_live: Bool, if True: Remove the live_data. Else remove processed data

        """
        # Query the database to remove data
        # WANRING! If we don't find something we WILL remove the data! Don't make the query
        # specific! We only check if any of the data is on this host (i.e. not get None from
        # the query below)
        rd = run_coll.find_one(
            {"number": number, "data.host": hostname if not delete_live else "daq"}
        )
        run_id = "%06d" % number

        if rd and any(
            ddoc.get("location", "").startswith(clean_from) for ddoc in rd.get("data", [])
        ):
            # Just for explicitness, this is where we want to end up. If we have a rundoc,
            # the data is registered and we don't have to do anything.
            return
        else:
            log_warning(
                f"remove_if_unregistered::\trun {number} is NOT registered "
                f"in the runDB but is stored on {hostname}",
                run_id=run_id,
                priority="error",
            )
            if not delete_live:
                # Check the local ebs disk for data.
                _remove_unregistered_run(clean_from, run_id, checked_db=True)
            else:
                # Check ceph for data associated to this run (which is apparently not in the
                # runDB)
                _remove_unregistered_run(ceph_folder, run_id, checked_db=True)

    def _remove_unregistered_run(base_folder, run_id, checked_db=False):
        """NB: The check that this run is not registered should be performed first!
        Deletes any folder from base_folder that matches run_id.

        :param base_folder: folder to check
        :param run_id: run_id to remove from folder
        :param checked_db: Bool if it is checked that this run in not in the database.

        """
        if not checked_db:
            log_warning(
                "remove_if_unregistered::\trogue ajax operations! Trying "
                f"to delete {run_id} from {hostname}",
                run_id=run_id,
                priority="fatal",
            )
            raise ValueError(
                "Only insert runs where for it is checked that it is "
                "not registered in the runs database and double checked."
            )
        log_warning(
            f"No data for {run_id} found! Double checking {base_folder}!",
            run_id=run_id,
            priority="warning",
        )
        deleted_data = False

        for folder in os.listdir(base_folder):
            if run_id in folder:
                log.info(f"Cleaning {base_folder + folder}")

                # Do a final check if we are not deleting essential data!
                # Do not disable this check! If you don't like it: make a smarter query
                rd = run_coll.find_one({"number": int(run_id)})
                n_live, n_rr = check_rundoc_for_live_and_raw(rd)

                if not (n_live or n_rr >= 1 + int("raw_records" in folder)):
                    message = (
                        f"Trying to delete {folder} but we only have "
                        f"{n_live} live- and {n_rr} raw_record-files in the "
                        "runs-database. This might be an essential copy of "
                        "the data!"
                    )
                    log_warning(message, run_id=run_id, priority="fatal")
                    raise ValueError(message)

                # OK, we still have live_data somewhere or we have raw_records (elsewhere)
                if args.production:
                    # Double check returns True automatically if not args.ask_confirm
                    if confirm(
                        f"Should we really move {os.path.join(base_folder, folder)} "
                        f"to {os.path.join(non_registered_folder, folder)}?"
                    ):
                        shutil.move(
                            os.path.join(base_folder, folder),
                            os.path.join(non_registered_folder, folder),
                        )
                else:
                    log.info(f"TEST\tmoving {base_folder + folder}")
                deleted_data = True

        if not deleted_data:
            message = f"No data registered on {hostname} for {run_id}"
            log_warning(message, priority="fatal")
            raise FileNotFoundError(message)

    def now(plus=0):
        """UTC timestamp."""
        return datetime.now(pytz.utc) + timedelta(seconds=plus)

    def confirm(question):
        """If --ask_confirm is specified, ask user to confirm to proceed.

        :return: bool

        """
        if not args.ask_confirm:
            return True
        answer = str(input(question + " (y/n): \n")).lower().strip()
        if answer in ("y", "n"):
            return answer == "y"
        else:
            confirm("please input (y/n)\n")


if __name__ == "__main__":
    main()



import numpy as np
import straxen
from straxen.plugins.defaults import DEFAULT_POSREC_ALGO

import strax

export, __all__ = strax.exporter()


@export
class EventPositions(strax.Plugin):
    """Computes the observed and corrected position for the main S1/S2 pairs in an event.

    For XENONnT data, it returns the FDC corrected positions of the
    default_reconstruction_algorithm. In case the fdc_map is given as a file (not through xedocs),
    then the coordinate system should be given as (x, y, z), not (x, y, drift_time).

    """

    depends_on = "event_basics"

    __version__ = "0.3.0"

    default_reconstruction_algorithm = straxen.URLConfig(
        default=DEFAULT_POSREC_ALGO, help="default reconstruction algorithm that provides (x,y)"
    )

    electron_drift_velocity = straxen.URLConfig(
        default="xedocs://electron_drift_velocities?attr=value&run_id=plugin.run_id&version=ONLINE",
        cache=True,
        help="Vertical electron drift velocity in cm/ns (1e4 m/ms)",
    )

    electron_drift_time_gate = straxen.URLConfig(
        default="xedocs://electron_drift_time_gates?attr=value&run_id=plugin.run_id&version=ONLINE",
        help="Electron drift time from the gate in ns",
        cache=True,
    )

    fdc_map = straxen.URLConfig(
        default="xedocs://fdc_maps"
        "?algorithm=plugin.default_reconstruction_algorithm&run_id=plugin.run_id"
        "&attr=map&scale_coordinates=plugin.coordinate_scale&version=ONLINE",
        infer_type=False,
        help="3D field distortion correction map path",
    )

    z_bias_map = straxen.URLConfig(
        default="itp_map://resource://XnT_z_bias_map_chargeup_20230329.json.gz"
        "?fmt=json.gz&method=RegularGridInterpolator",
        infer_type=False,
        help="Map of Z bias due to non uniform drift velocity/field",
    )

    def infer_dtype(self):
        dtype = []
        for j in "x y r".split():
            comment = f"Main interaction {j}-position, field-distortion corrected (cm)"
            dtype += [(j, np.float32, comment)]
            for s_i in [1, 2]:
                comment = (
                    f"Alternative S{s_i} interaction (rel. main S{3 - s_i}) {j}-position,"
                    " field-distortion corrected (cm)"
                )
                field = f"alt_s{s_i}_{j}_fdc"
                dtype += [(field, np.float32, comment)]

        for j in ["z"]:
            comment = "Interaction z-position, corrected to non-uniform drift velocity (cm)"
            dtype += [(j, np.float32, comment)]
            comment = (
                "Interaction z-position, corrected to non-uniform drift velocity, duplicated (cm)"
            )
            dtype += [(j + "_dv_corr", np.float32, comment)]
            for s_i in [1, 2]:
                comment = (
                    f"Alternative S{s_i} z-position (rel. main S{3 - s_i}), corrected to"
                    " non-uniform drift velocity (cm)"
                )
                field = f"alt_s{s_i}_z"
                dtype += [(field, np.float32, comment)]
                # values for corrected Z position
                comment = (
                    f"Alternative S{s_i} z-position (rel. main S{3 - s_i}), corrected to"
                    " non-uniform drift velocity, duplicated (cm)"
                )
                field = f"alt_s{s_i}_z_dv_corr"
                dtype += [(field, np.float32, comment)]

        naive_pos = []
        fdc_pos = []
        for j in "r z".split():
            naive_pos += [
                (
                    f"{j}_naive",
                    np.float32,
                    f"Main interaction {j}-position with observed position (cm)",
                )
            ]
            fdc_pos += [
                (
                    f"{j}_field_distortion_correction",
                    np.float32,
                    f"Correction added to {j}_naive for field distortion (cm)",
                )
            ]
            for s_i in [1, 2]:
                naive_pos += [
                    (
                        f"alt_s{s_i}_{j}_naive",
                        np.float32,
                        (
                            f"Alternative S{s_i} interaction (rel. main S{3 - s_i}) {j}-position"
                            " with observed position (cm)"
                        ),
                    )
                ]
                fdc_pos += [
                    (
                        f"alt_s{s_i}_{j}_field_distortion_correction",
                        np.float32,
                        f"Correction added to alt_s{s_i}_{j}_naive for field distortion (cm)",
                    )
                ]
        dtype += naive_pos + fdc_pos
        for s_i in [1, 2]:
            dtype += [
                (
                    f"alt_s{s_i}_theta",
                    np.float32,
                    f"Alternative S{s_i} (rel. main S{3 - s_i}) "
                    "interaction angular position (radians)",
                )
            ]

        dtype += [("theta", np.float32, f"Main interaction angular position (radians)")]
        return dtype + strax.time_fields

    def setup(self):
        self.coordinate_scales = [1.0, 1.0, -self.electron_drift_velocity]
        self.map = self.fdc_map

    def compute(self, events):
        result = {"time": events["time"], "endtime": strax.endtime(events)}

        # s_i == 0 indicates the main event,
        # while s_i != 0 means alternative S1 or S2 is used based on s_i value
        # while the other peak is the main one
        # (e.g., s_i == 1 means that the event is defined using altS1 and main S2)
        for s_i in [0, 1, 2]:
            # alt_sx_interaction_drift_time is calculated between main Sy and alternative Sx
            drift_time = (
                events["drift_time"] if not s_i else events[f"alt_s{s_i}_interaction_drift_time"]
            )
            z_obs = -self.electron_drift_velocity * drift_time
            xy_pos = "s2_" if s_i != 2 else "alt_s2_"
            orig_pos = np.vstack([events[f"{xy_pos}x"], events[f"{xy_pos}y"], z_obs]).T
            r_obs = np.linalg.norm(orig_pos[:, :2], axis=1)
            z_obs = z_obs + self.electron_drift_velocity * self.electron_drift_time_gate

            # apply z bias correction
            z_dv_delta = self.z_bias_map(np.array([r_obs, z_obs]).T, map_name="z_bias_map")
            corr_pos = np.vstack([events[f"{xy_pos}x"], events[f"{xy_pos}y"], z_obs - z_dv_delta]).T
            # apply r bias correction
            delta_r = self.map(corr_pos)
            with np.errstate(invalid="ignore", divide="ignore"):
                r_cor = r_obs + delta_r
                scale = np.divide(r_cor, r_obs, out=np.zeros_like(r_cor), where=r_obs != 0)

            pre_field = "" if s_i == 0 else f"alt_s{s_i}_"
            post_field = "" if s_i == 0 else "_fdc"
            result.update(
                {
                    f"{pre_field}x{post_field}": orig_pos[:, 0] * scale,
                    f"{pre_field}y{post_field}": orig_pos[:, 1] * scale,
                    f"{pre_field}r{post_field}": r_cor,
                    f"{pre_field}r_naive": r_obs,
                    f"{pre_field}r_field_distortion_correction": delta_r,
                    f"{pre_field}theta": np.arctan2(orig_pos[:, 1], orig_pos[:, 0]),
                    f"{pre_field}z_naive": z_obs,
                    # using z_dv_corr (z_obs - z_dv_delta) in agreement with the dtype description
                    # the FDC for z (z_cor) is found to be not reliable (see #527)
                    f"{pre_field}z": z_obs - z_dv_delta,
                    f"{pre_field}z_dv_corr": z_obs - z_dv_delta,
                }
            )
        return result
